/*
 *  arch/mips/emma3tl/common/irq_emma3tl.c
 *      This file defines the irq handler for EMMA3TL.
 *
 * Author: NEC Electronics Corporation
 *
 * Copyright (C) 2008 NEC Electronics Corporation
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *  This file is based on the arch/mips/ddb5xxx/ddb5477/irq_5477.c
 *
 *	Copyright 2001 MontaVista Software Inc.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * EMMA3TL defines 64 IRQs.
 *
 * This file exports one function:
 *	emma_irq_init(u32 irq_base);
 */

#include <linux/interrupt.h>
#include <linux/types.h>
#include <linux/ptrace.h>

#include <asm/debug.h>

#include <asm/emma3tl/emma3tl.h>
#ifdef CONFIG_E21
#include <asm/emma3tl/e21.h>
#endif

static int emma_irq_base = -1;
static int emma_sw_irq_base = -1;

void ll_emma_irq_enable(int);
void ll_emma_irq_disable(int);
void ll_emma_sw_irq_enable(int reg);
void ll_emma_sw_irq_disable(int reg);

static void emma_irq_enable(unsigned int irq)
{
	ll_emma_irq_enable(irq - emma_irq_base);
}

static void emma_irq_disable(unsigned int irq)
{
	ll_emma_irq_disable(irq - emma_irq_base);
}

static unsigned int emma_irq_startup(unsigned int irq)
{
	emma_irq_enable(irq);
	return 0;
}

#define	emma_irq_shutdown	emma_irq_disable

static void emma_irq_ack(unsigned int irq)
{
	/* disable interrupt - some handler will re-enable the irq
	 * and if the interrupt is leveled, we will have infinite loop
	 */
	ll_emma_irq_disable(irq - emma_irq_base);
}

static void emma_irq_end(unsigned int irq)
{
	if (!(irq_desc[irq].status & (IRQ_DISABLED | IRQ_INPROGRESS)))
		ll_emma_irq_enable(irq - emma_irq_base);
}

hw_irq_controller emma_irq_controller = {
	.typename = "emma_irq",
	.startup = emma_irq_startup,
	.shutdown = emma_irq_shutdown,
	.enable = emma_irq_enable,
	.disable = emma_irq_disable,
	.ack = emma_irq_ack,
	.end = emma_irq_end,
	.set_affinity = NULL	/* no affinity stuff for UP */
};

void emma_irq_init(u32 irq_base)
{
	u32 i;

	for (i = irq_base; i < irq_base + NUM_EMMA_IRQ; i++) {
		irq_desc[i].status = IRQ_DISABLED;
		irq_desc[i].action = NULL;
		irq_desc[i].depth = 1;
		irq_desc[i].chip = &emma_irq_controller;
	}

	emma_irq_base = irq_base;
}

void ll_emma_irq_enable(int emma_irq)
{
	u32 reg_value;
	u32 reg_bitmask;
	u32 reg_index;

	reg_index = EMMA_BHIF_INT0_EN_0
	    + (EMMA_BHIF_INT0_EN_1 - EMMA_BHIF_INT0_EN_0)
	    * (emma_irq / 32);
	reg_value = emma_in32(reg_index);
	reg_bitmask = 0x1 << (emma_irq % 32);
	db_assert((reg_value & reg_bitmask) == 0);
	emma_out32(reg_index, reg_value | reg_bitmask);
}

void ll_emma_irq_disable(int emma_irq)
{
	u32 reg_value;
	u32 reg_bitmask;
	u32 reg_index;

	reg_index = EMMA_BHIF_INT0_EN_0
	    + (EMMA_BHIF_INT0_EN_1 - EMMA_BHIF_INT0_EN_0)
	    * (emma_irq / 32);
	reg_value = emma_in32(reg_index);
	reg_bitmask = 0x1 << (emma_irq % 32);
	db_assert((reg_value & reg_bitmask) != 0);
	emma_out32(reg_index, reg_value & ~reg_bitmask);
}

static void emma_sw_irq_enable(unsigned int irq)
{
	ll_emma_sw_irq_enable(irq - emma_sw_irq_base);
}

static void emma_sw_irq_disable(unsigned int irq)
{
	ll_emma_sw_irq_disable(irq - emma_sw_irq_base);
}

static unsigned int emma_sw_irq_startup(unsigned int irq)
{
	emma_sw_irq_enable(irq);
	return 0;
}

#define emma_sw_irq_shutdown emma_sw_irq_disable

static void emma_sw_irq_ack(unsigned int irq)
{
	ll_emma_sw_irq_disable(irq - emma_sw_irq_base);
}

static void emma_sw_irq_end(unsigned int irq)
{
	if (!(irq_desc[irq].status & (IRQ_DISABLED | IRQ_INPROGRESS)))
		ll_emma_sw_irq_enable(irq - emma_sw_irq_base);
}

hw_irq_controller emma_sw_irq_controller = {
	.typename = "emma_sw_irq",
	.startup = emma_sw_irq_startup,
	.shutdown = emma_sw_irq_shutdown,
	.enable = emma_sw_irq_enable,
	.disable = emma_sw_irq_disable,
	.ack = emma_sw_irq_ack,
	.end = emma_sw_irq_end,
	.set_affinity = NULL,
};

void emma_sw_irq_init(u32 irq_base)
{
	u32 i;

	for (i = irq_base; i < irq_base + NUM_EMMA_IRQ_SW; i++) {
		irq_desc[i].status = IRQ_DISABLED;
		irq_desc[i].action = NULL;
		irq_desc[i].depth = 2;
		irq_desc[i].chip = &emma_sw_irq_controller;
	}

	emma_sw_irq_base = irq_base;
}

void ll_emma_sw_irq_enable(int irq)
{
	u32 reg;

	db_assert(irq >= 0);
	db_assert(irq < NUM_EMMA_IRQ_SW);

	reg = emma_in32(EMMA_BHIF_SW_INT0_EN);
	reg |= 1 << irq;
	emma_out32(EMMA_BHIF_SW_INT0_EN, reg);
}

void ll_emma_sw_irq_disable(int irq)
{
	u32 reg;

	db_assert(irq >= 0);
	db_assert(irq < 32);

	reg = emma_in32(EMMA_BHIF_SW_INT0_EN);
	reg &= ~(1 << irq);
	emma_out32(EMMA_BHIF_SW_INT0_EN, reg);
}

/*
 * the first level int-handler will jump here if it is a emma irq
 */
asmlinkage void emma_irq_dispatch(struct pt_regs *regs)
{
	u32 intStatus;
	u32 bitmask;
	u32 i;

	intStatus = emma_in32(EMMA_BHIF_INT_ST_0)
	    & emma_in32(EMMA_BHIF_INT0_EN_0);

#ifdef EMMA_SW_CASCADE
	if (intStatus &
	    (1 << ((EMMA_SW_CASCADE - EMMA_IRQ_BASE) & (32 - 1)))) {
		u32 swIntStatus;
		swIntStatus = emma_in32(EMMA_BHIF_SW_INT0)
		    & emma_in32(EMMA_BHIF_SW_INT0_EN);
		for (i = 0, bitmask = 1; i < 32; i++, bitmask <<= 1) {
			if (swIntStatus & bitmask) {
				do_IRQ(EMMA_SW_IRQ_BASE + i, regs);
				return;
			}
		}
	}
	/* ignoring SW interrupt */
	intStatus &= ~(1UL << (EMMA_SW_CASCADE - EMMA_IRQ_BASE));
#endif

	for (i = 0, bitmask = 1; i < 32; i++, bitmask <<= 1) {
		if (intStatus & bitmask) {
			do_IRQ(EMMA_IRQ_BASE + i, regs);
			return;
		}
	}

	intStatus = emma_in32(EMMA_BHIF_INT_ST_1)
	    & emma_in32(EMMA_BHIF_INT0_EN_1);

	for (i = 32, bitmask = 1; i < 64; i++, bitmask <<= 1) {
		if (intStatus & bitmask) {
			do_IRQ(EMMA_IRQ_BASE + i, regs);
			return;
		}
	}

	/* Dispatch from IRQ64 to 95 */
	intStatus = emma_in32(EMMA_BHIF_INT_ST_2)
	    & emma_in32(EMMA_BHIF_INT0_EN_2);

	for (i = 64, bitmask = 1; i < 96; i++, bitmask <<= 1) {
		if (intStatus & bitmask) {
			do_IRQ(EMMA_IRQ_BASE + i, regs);
			return;
		}
	}
}

