/*
 * i2c-emma.c
 *
 * I2C adapter for the EMMA I2C bus access.
 *
 * Author: NEC Electronics Corporation
 *
 * Copyright (C) 2007-2008 NEC Electronics Corporation
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */


#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/i2c.h>
#include <linux/i2c-id.h>
#include <linux/init.h>
#include <linux/time.h>
#include <linux/sched.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/interrupt.h>
#include <linux/platform_device.h>
#include "i2c-emma3tl.h"

#include <linux/io.h>

#define MAX_NUM_DEVS	3
#define TOGGLE_NUM		250 /* I2C復帰処理時に発生させるクロック回数 Mod by HASE@10021801 was 9 */

#define NOACK			0x1
#define T_ABRT			0x2
#define I2C_ABRT_LOST	0x4
#define I2C_ERROR		0x8

/* Following are the default values for the module parameters */
/* I2C#0 default Fast Mode = 400 KHz,Standard Mode = 100 KHz */
ulong clock0 = 0; /* was 1 I2C#0 には OFDM と Tuner がぶら下がっていて Standard Mode で十分なので 変更 by HASE@09112703 */
/* I2C#1 default Fast Mode = 400 KHz,Standard Mode = 100 KHz */
ulong clock1 = 0; /* was 1 I2C#1 にぶら下がっている 音多IC が Standard Mode にしか対応していないので 変更 by HASE@09082001 */
/* I2C#2 default Fast Mode = 400 KHz,Standard Mode = 100 KHz */
ulong clock2 = 1;
u32 clock[2] = {STANDARD_MODE, FAST_MODE};

struct i2c_msg_desc {
	__u16   num;
	__u16   pos;
};

struct emma_i2c {
	struct i2c_msg		*msgs;
	int			msg_num;
	struct i2c_msg_desc	tx_desc;
	struct i2c_msg_desc	rx_desc;
	unsigned short		rx_remainder;
	int			completed;

	int				enable;
	wait_queue_head_t	wait;

	struct i2c_adapter	adap;
	u32			ptop_base;
	u32			pbrd_base;
	u32			clock;
	u32			state;
	int			id;
	unsigned int		irq;
};

static void i2c_emma_setreg(struct emma_i2c *i2c, u32 ctl, u32 val)
{
	u32 addr;

	addr = i2c->ptop_base + ctl;
	writel(val, (void *)addr);
}

static u32 i2c_emma_getreg(struct emma_i2c *i2c, u32 ctl)
{
	u32 addr;

	addr = i2c->ptop_base + ctl;
	return readl((void *)addr);
}

static void i2c_emma_enable(struct emma_i2c *i2c)
{
	int x, cnt;

#if 1 /* Activity Status時に, I2CコントローラをEnable/Disableにしてはいけません Add by HASE@10021801 */
	// UserManualの 6.16.2.1.25. I2C Enable の ACTIVITYビットについての説明を参照
	if ((x=i2c_emma_getreg(i2c, IC_STATUS)) & ACTIVITY_STATUS) {
		
	//	printk(KERN_ALERT "wng : %s() Activity Status [%s]\n", __FUNCTION__,
	//				i2c->adap.dev.bus_id);
		
		/* Activity Statusならば Inactivity Statusになるまで待ち */
		for (cnt = 0; cnt < 30; cnt++) {
			mdelay(10);
			if (!((x=i2c_emma_getreg(i2c, IC_STATUS)) & ACTIVITY_STATUS)) {
				break;
			}
		}
		// 300msec待ってもダメならば, 強制的に Enable とする
		if (30 <= cnt) {
			printk(KERN_ALERT "err : %s() disable error [%s]\n", __FUNCTION__,
						i2c->adap.dev.bus_id);
		}
	}
#endif

	i2c_emma_setreg(i2c, IC_ENABLE, ENABLE);
}

static int i2c_emma_disable(struct emma_i2c *i2c)
{
	int ret = 0;
	int x, cnt;

#if 1 /* Activity Status時に, I2CコントローラをEnable/Disableにしてはいけません Add by HASE@10021801 */
	// UserManualの 6.16.2.1.25. I2C Enable の ACTIVITYビットについての説明を参照
	if ((x=i2c_emma_getreg(i2c, IC_STATUS)) & ACTIVITY_STATUS) {
		
	//	printk(KERN_ALERT "wng : %s() Activity Status [%s]\n", __FUNCTION__,
	//				i2c->adap.dev.bus_id);
		
		/* Activity Statusならば Inactivity Statusになるまで待ち */
		for (cnt = 0; cnt < 30; cnt++) {
			mdelay(10);
			if (!((x=i2c_emma_getreg(i2c, IC_STATUS)) & ACTIVITY_STATUS)) {
				break;
			}
		}
		// 300msec待ってもダメならば, 強制的に Disable とする
		if (30 <= cnt) {
			printk(KERN_ALERT "err : %s() disable error [%s]\n", __FUNCTION__,
						i2c->adap.dev.bus_id);
			ret = 1;
		}
	}
#endif

	i2c_emma_setreg(i2c, IC_ENABLE, 0x0);

	if (i2c_emma_getreg(i2c, IC_ENABLE_STATUS) & IC_EN) {
		i2c->state |= I2C_ERROR;
		ret = 1;
	}

	return ret;
}

/* I2C_ABRT_LOST 発生時に本関数にて 復帰処理が行われます              *
 * I2C_ABRT_LOST とは スレーブの石がノイズ等で誤ってACKを返したままの *
 * (SDA=L)の状態で その間はI2CコントローラがIDLE状態になり, コマンド  *
 * を受け付けてくれなくなる. その際は I2C_ABRT割り込み発生によって    *
 * I2C_ABRT_LOST が検知され 本関数がCallされて復帰を図る              *
 * (UserManual 3.1. 使用上の注意点 を参照)                            */
static void i2c_emma_recover(struct emma_i2c *i2c)
{
	int i;
	int delay;
	u32 speed;
	u32 fail_safe;
	u32 reg[2];

	BUG_ON(i2c->id >= MAX_NUM_DEVS);

	speed = i2c_emma_getreg(i2c, IC_CON) & IC_SPEED;
	switch (speed) {
	case STANDARD_MODE:
		delay = 5; /* 100kbpsなので 1/2クロック時間は (1000000 / (100*1000)) / 2 = 5usec */
		break;
	case FAST_MODE:
		delay = 3;
		break;
	default:
		delay = 1;
	}

	reg[0] = readl((void *)(i2c->pbrd_base + I2C_FAIL_SAFE));
	reg[1] = readl((void *)(i2c->pbrd_base + I2C_SDA_DELAY));

	/* toggle mode set */
	fail_safe = 1 << (i2c->id * 8);
	writel(fail_safe, (void *)(i2c->pbrd_base + I2C_FAIL_SAFE));

	/* SCL toggle */
	/* (クロック線を上げ下げして スレーブの石のI2C処理を進めてあげて 復帰を図る) */
	for (i = 0; i < TOGGLE_NUM; i++) {
		writel(reg[1] & ~fail_safe,
			(void *)(i2c->pbrd_base + I2C_SDA_DELAY));
		udelay(delay); /* 1/2クロック時間待ち */
		writel(reg[1] | fail_safe,
			(void *)(i2c->pbrd_base + I2C_SDA_DELAY));
		udelay(delay);  /* 1/2クロック時間待ち */
	}

	/* SDA toggle */
	writel(reg[1] & ~(fail_safe << 1),
		(void *)(i2c->pbrd_base + I2C_SDA_DELAY));
	writel(reg[1] | (fail_safe << 1),
		(void *)(i2c->pbrd_base + I2C_SDA_DELAY));
	writel(reg[1] & ~(fail_safe << 1),
		(void *)(i2c->pbrd_base + I2C_SDA_DELAY));

	i2c_emma_disable(i2c);

	/* regular mode set */
	writel(reg[0], (void *)(i2c->pbrd_base + I2C_FAIL_SAFE));
	writel(reg[1], (void *)(i2c->pbrd_base + I2C_SDA_DELAY));
}

static void i2c_emma_set_slaveaddr(struct emma_i2c *i2c, struct i2c_msg *msg)
{
	u32 addr;

	if (msg->flags & I2C_M_TEN) {
		addr = 0xf0 | ((msg->addr >> 7) & 0x03);
		printk(KERN_ALERT
			"err : %s() 10 bit addresses are not supported in this driver.\n", __FUNCTION__);
	} else
		addr = msg->addr & 0x7f;

	i2c_emma_setreg(i2c, IC_TAR, addr);
}

static void i2c_emma_set_clock(struct emma_i2c *i2c, u32 val)
{
	u32 icr;

	icr = i2c_emma_getreg(i2c, IC_CON) & ~IC_SPEED;
	i2c_emma_setreg(i2c, IC_CON, icr | val);
}

static u32 i2c_emma_getclearints(struct emma_i2c *i2c)
{
	u32 intStatus;

	intStatus = i2c_emma_getreg(i2c, IC_INTR_STAT);

	if (intStatus & INTR_RX_UNDER)
		i2c_emma_getreg(i2c, IC_CLR_RX_UNDER);
	if (intStatus & INTR_RX_OVER)
		i2c_emma_getreg(i2c, IC_CLR_RX_OVER);
	if (intStatus & INTR_TX_OVER)
		i2c_emma_getreg(i2c, IC_CLR_TX_OVER);
	if (intStatus & INTR_RD_REQ)
		i2c_emma_getreg(i2c, IC_CLR_RD_REQ);
	if (intStatus & INTR_TX_ABRT)
		i2c_emma_getreg(i2c, IC_CLR_TX_ABRT);
	if (intStatus & INTR_RX_DONE)
		i2c_emma_getreg(i2c, IC_CLR_RX_DONE);
	if (intStatus & INTR_ACTIVITY)
		i2c_emma_getreg(i2c, IC_CLR_ACTIVITY);
	if (intStatus & INTR_STOP_DET)
		i2c_emma_getreg(i2c, IC_CLR_STOP_DET);
	if (intStatus & INTR_START_DET)
		i2c_emma_getreg(i2c, IC_CLR_START_DET);
	if (intStatus & INTR_GEN_CALL)
		i2c_emma_getreg(i2c, IC_CLR_GEN_CALL);

	return intStatus;
}

/*
 * We are protected by the adapter bus semaphore.
 */
static int i2c_emma_do_xfer(struct emma_i2c *i2c, struct i2c_msg *msgs, int num)
{
	long timeout;
	int i;
	u32 intrmask;
	int ret = 0;

	if (i2c->enable) {
		printk(KERN_ALERT "err : %s() I2C bus is used", __FUNCTION__);
		goto err_out;
	}

	/* EMMA2TLのI2CコントローラのFIFOバッファは64段(8bitx64)しかないので
	   一度の通信が64byteを超えるI2C読み書きは控えてください */
	if(64 < msgs->len){
		printk(KERN_ALERT "wng : %s() [%d]byte is too long\n", __FUNCTION__, msgs->len);
	}

	i2c->msgs = msgs;
	i2c->msg_num = num;
	i2c->completed = 0;

	/*
	 * Search for the 1st read msg.
	 * Even if there is no read msg, just leave it.
	 */
	for (i = 0; i < i2c->msg_num; i++)
		if (i2c->msgs[i].flags & I2C_M_RD)
			break;
	/* rx descriptor */
	i2c->rx_desc.num = i;
	i2c->rx_desc.pos = 0;
	i2c->rx_remainder = 0;

	/* tx descriptor */
	i2c->tx_desc.num = 0;
	i2c->tx_desc.pos = 0;

	/* RX_FULL,TX_EMPTY interrupt mask clear */
	intrmask = i2c_emma_getreg(i2c, IC_INTR_MASK);
	intrmask &= ~(INTR_RX_FULL | INTR_TX_EMPTY);
	i2c_emma_setreg(i2c, IC_INTR_MASK, intrmask);
	i2c_emma_setreg(i2c, IC_DATA_CMD, 0); // writeモードに戻しておくこと！

	/* interrupt status clear */
	i2c_emma_getclearints(i2c);

	/* target address set */
	i2c_emma_set_slaveaddr(i2c, i2c->msgs);

	intrmask = i2c_emma_getreg(i2c, IC_INTR_MASK);
	intrmask |= (INTR_STOP_DET | INTR_RX_FULL |
			INTR_TX_EMPTY | INTR_TX_ABRT);
	i2c_emma_setreg(i2c, IC_INTR_MASK, intrmask);

	i2c->enable = 1;

	i2c_emma_enable(i2c);

	/*
	 * The rest of the processing occurs in the interrupt handler.
	 */
	timeout = wait_event_timeout(i2c->wait, i2c->completed == 1, HZ * 3 / 10);
		/* Mod by HASE@10021801 was HZ * 5                       ↑                 *
		 * HZ * 5 って... 5秒なんてありえません Standard Mode(100kbps)でも100msecを *
		 * 超えることはまずない(実動でも確認済み) 余裕を見て3倍の 300msec にする    */

#if 1 /* ABORT 発生時には クロック線を上げ下げして スレーブの石のI2C処理を進めてあげて 復帰を図る Add by HASE@10021801 */
	if (i2c->state & T_ABRT) {
		printk(KERN_ALERT "wng : %s() NOACK or I2C_ABRT_LOST or T_ABRT", __FUNCTION__);
		i2c_emma_recover(i2c);
		i2c->enable = 0;
		goto err_out;
	}
#endif

	if (timeout == 0) {
		printk(KERN_ALERT "err : %s() timeout", __FUNCTION__);
#if 1 /* クロック線を上げ下げして スレーブの石のI2C処理を進めてあげて 復帰を図る Add by HASE@10021801 */
		i2c_emma_recover(i2c);
#endif
		i2c->enable = 0;
		goto err_out;
	}

	if (i2c_emma_disable(i2c)) {
		printk(KERN_ALERT "err : %s() disable error", __FUNCTION__);
		i2c->enable = 0;
		goto err_out;
	}

	i2c->enable = 0;

	if (((i2c->tx_desc.num < i2c->msg_num) || (i2c->rx_remainder)) &&
	    ((i2c->state & T_ABRT) == 0)) {
		if(i2c->rx_remainder){
			printk(KERN_ERR "wng : %s() Can't receive data", __FUNCTION__);
		}
		else{
			printk(KERN_ERR "wng : %s() Tx FIFO underrun", __FUNCTION__);
		}
		i2c->state |= I2C_ERROR;
		ret = -EAGAIN; goto err_out;
	}

	return i2c->tx_desc.num;

err_out:

	printk(" [%s] SlvAdr[0x%02x] [%08x]\n", i2c->adap.dev.bus_id, (msgs->addr << 1) | ((msgs->flags & I2C_M_RD) ? 1 : 0), i2c->state);

	return ret;
}

/*
 * i2c_emma_master_complete - complete the message and wake up.
 */
static inline void i2c_emma_master_complete(struct emma_i2c *i2c)
{
	i2c_emma_setreg(i2c, IC_INTR_MASK, 0x00000000);
	wake_up(&i2c->wait);
}

/*
 * EMMA I2C adapter handler
 */
static irqreturn_t i2c_emma_handler(
			int this_irq, void *dev_id, struct pt_regs *regs)
{
	struct emma_i2c *i2c = dev_id;
	u32 abrtsource;
	u32 intrstatus;
	u32 intrmask;
	struct i2c_msg *tx_msg = NULL, *rx_msg = NULL;
	int i;

	abrtsource = i2c_emma_getreg(i2c, IC_TX_ABRT_SOURCE);
	intrstatus = i2c_emma_getclearints(i2c);

	/* Transmit Abort */
	if (intrstatus & INTR_TX_ABRT) {

		if (abrtsource & (ABRT_GCALL_NOACK | ABRT_TXDATA_NOACK |
				  ABRT_10ADDR2_NOACK | ABRT_10ADDR1_NOACK |
				  ABRT_7B_ADDR_NOACK))
			i2c->state |= NOACK;

		/* Master has lost arbitrat */
		if (abrtsource & ABR_LOST) {
			i2c->state |= I2C_ABRT_LOST;
			/* 後で i2c_emma_do_xfer() 内で i2c_emma_recover() にて復帰を図ります */
		}
		i2c->state |= T_ABRT;
		goto err_out;
	}

	if (i2c->rx_desc.num < i2c->msg_num)
		rx_msg = &i2c->msgs[i2c->rx_desc.num];
	if (i2c->tx_desc.num < i2c->msg_num)
		tx_msg = &i2c->msgs[i2c->tx_desc.num];

	while (rx_msg && i2c->rx_remainder &&
	       (i2c_emma_getreg(i2c, IC_STATUS) & RFNE)) {

		rx_msg->buf[i2c->rx_desc.pos] =
			(u8)(i2c_emma_getreg(i2c, IC_DATA_CMD) & IC_DAT);
		/*
		 * Update rx_desc
		 *
		 *
		 */
		i2c->rx_desc.pos++;
		i2c->rx_remainder--;

		if (i2c->rx_desc.pos >= rx_msg->len) {
			i2c->rx_desc.num++;
			i2c->rx_desc.pos = 0;
			if (i2c->rx_desc.num < i2c->msg_num) {
				/* Search the next read msg */
				for (i = i2c->rx_desc.num;
					i < i2c->msg_num; i++)
					if (i2c->msgs[i].flags & I2C_M_RD)
						break;
				i2c->rx_desc.num = i;
				rx_msg = &i2c->msgs[i2c->rx_desc.num];
			} else
				/* quit data reception */
				rx_msg = NULL;
		}
	}

	while (tx_msg && (i2c->tx_desc.pos < tx_msg->len) &&
	       (i2c_emma_getreg(i2c, IC_STATUS) & TFNF)) {

		if (tx_msg->flags & I2C_M_RD) {
			i2c_emma_setreg(i2c, IC_DATA_CMD, IC_CMD);
			i2c->rx_remainder++;
		} else
			i2c_emma_setreg(i2c, IC_DATA_CMD,
					tx_msg->buf[i2c->tx_desc.pos]);
		/*
		 * Update tx_desc
		 * 1. increment tx_desc.pos
		 * 2. check tx_desc.pos wrap around
		 * 3. check msg num
		 */
		i2c->tx_desc.pos++;
		if (i2c->tx_desc.pos >= tx_msg->len) {
			i2c->tx_desc.num++;
			i2c->tx_desc.pos = 0;
			/*
			 * Move onto the next msg, or quit transmission
			 */
			if (i2c->tx_desc.num < i2c->msg_num)
				tx_msg = &i2c->msgs[i2c->tx_desc.num];
			else {
				/*
				 * Now all data have been written into Tx FIFO.
				 * There is no need to care about the TX_EMPTY
				 * interrupt, let's disable it.
				 */
				intrmask = i2c_emma_getreg(i2c, IC_INTR_MASK);
				intrmask &= ~INTR_TX_EMPTY;
				i2c_emma_setreg(i2c, IC_INTR_MASK, intrmask);
				break;
			}
		}
	}

	if (intrstatus & INTR_STOP_DET) {
err_out:
		i2c->completed = 1;
		i2c_emma_master_complete(i2c);
	}

	return IRQ_HANDLED;
}

static int i2c_emma_xfer(
		struct i2c_adapter *adap, struct i2c_msg msgs[], int num)
{
	struct emma_i2c *i2c;
	int ret = 0;
	int i;

	i2c = adap->algo_data;

	for (i = adap->retries; i >= 0; i--) {
		i2c->state = 0x00000000;
		ret = i2c_emma_do_xfer(i2c, msgs, num);
		if (ret == 0) {
			ret = -EREMOTEIO;
			goto out;
		} else if ((i2c->state & I2C_ERROR) == 0)
			goto out;
	}
out:
	return ret;
}

static void i2c_emma_hwinit(struct emma_i2c *i2c)
{

	u32 reg;

	i2c_emma_disable(i2c);
	if (i2c->id < MAX_NUM_DEVS) {
		reg = readl((void *)(i2c->pbrd_base + I2C_SDA_DELAY));
		reg &= ~(0xff << i2c->id * 8);
		reg |= (0x19 << i2c->id * 8);
		writel(reg, (void *)(i2c->pbrd_base + I2C_SDA_DELAY));
	}
	i2c_emma_setreg(i2c, IC_RX_TL, 0x0000003F);
	i2c_emma_setreg(i2c, IC_TX_TL, 0x0000003F);
	i2c_emma_setreg(i2c, IC_CON, 0x00000065);
	i2c_emma_set_clock(i2c, i2c->clock);
#if 1 /*I2Cバスのクロックを正しいクロックへ修正 by HASE@09081901
        (I2CコントローラのクロックはCTOP_STCCLK_SELのIICCLK_SEL[29:28]にて 81.92MHz に設定されています.
         E3TL_ClockDistributor_v1.00.pdf および レジスタ値 を参照すること)*/
	i2c_emma_setreg(i2c, IC_SS_SCL_HCNT, 0x00000181); /*Standard Mode : 100(↓のレジスタ値だと73)KHz*/
	i2c_emma_setreg(i2c, IC_SS_SCL_LCNT, 0x000001c0); 
	i2c_emma_setreg(i2c, IC_FS_SCL_HCNT, 0x0000003F); /*Fast Mode : 400(↓のレジスタ値だと340)KHz*/
	i2c_emma_setreg(i2c, IC_FS_SCL_LCNT, 0x0000007A); 
#else
	i2c_emma_setreg(i2c, IC_SS_SCL_HCNT, 0x000001D9);
	i2c_emma_setreg(i2c, IC_SS_SCL_LCNT, 0x00000212);
	i2c_emma_setreg(i2c, IC_FS_SCL_HCNT, 0x00000045);
	i2c_emma_setreg(i2c, IC_FS_SCL_LCNT, 0x0000007A);
#endif
	i2c_emma_setreg(i2c, IC_INTR_MASK, 0x00000000);
}

static int algo_control(struct i2c_adapter *adapter,
			unsigned int cmd, unsigned long arg)
{
	return 0;
}

static u32 emma_func(struct i2c_adapter *adap)
{
	return	I2C_FUNC_I2C | I2C_FUNC_SMBUS_WRITE_BYTE |
		I2C_FUNC_SMBUS_READ_BYTE |
		I2C_FUNC_SMBUS_WRITE_BYTE_DATA |
		I2C_FUNC_SMBUS_READ_BYTE_DATA |
		I2C_FUNC_SMBUS_WRITE_WORD_DATA |
		I2C_FUNC_SMBUS_READ_WORD_DATA |
		I2C_FUNC_SMBUS_READ_BLOCK_DATA |
		I2C_FUNC_SMBUS_WRITE_BLOCK_DATA |
		I2C_FUNC_SMBUS_READ_I2C_BLOCK  |
		I2C_FUNC_SMBUS_WRITE_I2C_BLOCK |
		I2C_FUNC_SMBUS_READ_I2C_BLOCK_2 |
		I2C_FUNC_SMBUS_WRITE_I2C_BLOCK_2;
}

static struct i2c_algorithm i2c_emma_algorithm = {
	.master_xfer		= i2c_emma_xfer,
	.smbus_xfer		= NULL,
	.slave_send		= NULL,
	.slave_recv		= NULL,
	.algo_control		= algo_control,
	.functionality		= emma_func,
};

static int __devinit i2c_emma_probe(struct device *dev)
{
	int ret;
	u32 cl = 0;
	struct resource *r;
	struct emma_i2c *i2c = NULL;
	struct platform_device *pdev = to_platform_device(dev);
	i2c = kzalloc(sizeof(struct emma_i2c), GFP_KERNEL);
	if (i2c == NULL) {
		ret = -ENOMEM;
		goto out;
	}
	i2c->adap.algo_data = i2c;
	strcpy(i2c->adap.name, dev->bus_id);
	i2c->adap.algo = &i2c_emma_algorithm;
	i2c->adap.retries = 5;
	i2c->enable = 0;
	init_waitqueue_head(&i2c->wait);
	dev_set_drvdata(dev, i2c);
	r = platform_get_resource(pdev, 0, 0);
	i2c->ptop_base = r->start;
	i2c->irq = platform_get_irq(pdev, 0);
	i2c->id = pdev->id;
	r = platform_get_resource_byname(pdev, 0, "pbrd");
	i2c->pbrd_base = r->start;

	BUG_ON(i2c->id >= MAX_NUM_DEVS);

	switch (i2c->id) {
	case 0:
		cl = clock0;
		break;
	case 1:
		cl = clock1;
		break;
	case 2:
		cl = clock2;
		break;
	}

	if (cl == 0)
		i2c->clock = clock[0];
	else if (cl == 1)
		i2c->clock = clock[1];
	else {
		printk(KERN_INFO "I2C: Failed to set clock\n");
		return -1;
	}

	ret = request_irq(i2c->irq, i2c_emma_handler, 0, dev->bus_id, i2c);
	if (ret < 0)
		goto out_free;
	ret = i2c_add_adapter(&i2c->adap);
	if (ret < 0) {
		printk(KERN_INFO "I2C: Failed to add bus\n");
		free_irq(i2c->irq, i2c);
		goto out_free;
	}
	printk(KERN_INFO "I2C: %s: EMMA I2C adapter\n", i2c->adap.dev.bus_id);
	i2c_emma_hwinit(i2c);

	return 0;

 out_free:
	kfree(i2c);

 out:
	return ret;
}

static int __devexit i2c_emma_remove(struct device *dev)
{
	struct emma_i2c *i2c = dev_get_drvdata(dev);

	if (i2c) {
		disable_irq(i2c->irq);
		free_irq(i2c->irq, i2c);
		i2c_del_adapter(&i2c->adap);
		kfree(i2c);
	}
	return 0;
}

static struct device_driver i2c_emma_driver = {
	.bus		= &platform_bus_type,
	.name		= "emma-iic",
	.probe		= i2c_emma_probe,
	.remove		= i2c_emma_remove,
};

static int __init i2c_adap_emma_init(void)
{
	return driver_register(&i2c_emma_driver);
}

static void i2c_adap_emma_exit(void)
{
	driver_unregister(&i2c_emma_driver);
}

MODULE_AUTHOR("NEC Electronics Corporation <www.necel.com>");
MODULE_DESCRIPTION("I2C-Bus adapter routines for EMMA I2C bus adapter");
MODULE_LICENSE("GPL v2");
MODULE_PARM_DESC(clock0,
		"I2C#0 CLOCK - 0 STANDARD_MODE; 1 FAST_MODE");
module_param(clock0, ulong, S_IRUGO);
MODULE_PARM_DESC(clock1,
		"I2C#1 CLOCK - 0 STANDARD_MODE; 1 FAST_MODE");
module_param(clock1, ulong, S_IRUGO);
MODULE_PARM_DESC(clock2,
		"I2C#2 CLOCK - 0 STANDARD_MODE; 1 FAST_MODE");
module_param(clock2, ulong, S_IRUGO);

module_init(i2c_adap_emma_init);
module_exit(i2c_adap_emma_exit);

