/*
 * $Id: physmap.c,v 1.39 2005/11/29 14:49:36 gleixner Exp $
 *
 * Normal mappings of chips in physical memory
 *
 * Copyright (C) 2003 MontaVista Software Inc.
 * Author: Jun Sun, jsun@mvista.com or jsun@junsun.net
 *
 * 031022 - [jsun] add run-time configure and partition setup
 */

#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/device.h>
#include <linux/platform_device.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/map.h>
#include <linux/mtd/partitions.h>
#include <linux/mtd/physmap.h>
#include <asm/io.h>

struct physmap_flash_info {
	struct mtd_info		*mtd;
	struct map_info		map;
	struct resource		*res;
#ifdef CONFIG_MTD_PARTITIONS
	int			nr_parts;
	struct mtd_partition	*parts;
#endif
};

#if 1 /* 2009.10.30 Add for Bootloaderで仕掛けたNOR Flash->DDRへのDMA転送 と LinuxのNOR FlashへのWrite 排他 by hiroshi.hasegawa@pixela.co.jp */

/*
 * EMMA2TL DMA Controller レジスタ
 */
#define EMMAxTL_DMAC_REG_BASE_ADR	(0xb0009000LU)
#define	EMMAxTL_DMAC_LIST_RST		(EMMAxTL_DMAC_REG_BASE_ADR + 0x080)
#define	EMMAxTL_DMAC_LIST_ENABLE	(EMMAxTL_DMAC_REG_BASE_ADR + 0x088)
#define	EMMAxTL_DMAC_0_LIST0		(EMMAxTL_DMAC_REG_BASE_ADR + 0x100)
#define SetReg( adr, dat )	(*(volatile uint32_t *)(adr) = dat )
#define GetReg( adr )		(*(volatile uint32_t *)(adr))

/* DMAコントローラ LISTレジスタ(E3TL_DMAController_v0.01.pdfを参照) */
typedef struct {
	unsigned int src;
	unsigned int dst;
	int          size;
	unsigned int config;
	void         *next;
	unsigned int skip_size;
	unsigned int list_load;
	unsigned int reserve;
} dma_list_t; /* Total 0x20 byte になるように */

/*
	void physmap_flash_pause_dma_by_bootloader(int pause)
	Bootloaderにて仕掛けた NOR Flash -> DDR へのDMA転送の 一時停止/解除 を行う
	(詳細は \\Odessa2\HAV\06_TL_FullARIB\a0_技術調査\起動時間短縮 を参照)
	
	※※※注意※※※
	本関数は xxx_pause_xxx(1) ～ xxx_pause_xxx(0) までを 排他して使用する必要があり
	利用するモジュールの責任で 排他を保障して使用すること
	また DMA 1,2chは 無視しています. BootloaderにてDMA転送を仕掛ける場合は DMA 0ch を
	利用すること(複数の転送がある場合には 転送LIST を利用して 0ch のみで転送を行う)
*/
void physmap_flash_pause_dma_by_bootloader(int pause)
{
	static int _dma0_not_active = 0;
	
/*
	static int _last_pause = 0;
	if(_last_pause == pause){
		printk("<0>--------> Pause call twice %d, %d\n", _last_pause, pause);
	}
	else{
		_last_pause = pause;
	}
*/
	
	/* DMA転送が未使用 or すでに完了ならば 不要 */
	if(!_dma0_not_active){
		volatile dma_list_t *reg;
		
		/* DMA転送の一時停止/再開 */
		reg = (dma_list_t *)EMMAxTL_DMAC_0_LIST0;
		/* DMA転送を一時停止 */
		if(pause){            /* ↓[27] dst_um, [26] src_um */     /*   ただし DDR<->DDR間の転送でなく */
			if(((reg->config & 0x0C000000LU) != 0x0C000000LU)      /* ←(ROM<->DDR間の転送であり) かつ */
			&& (GetReg(EMMAxTL_DMAC_LIST_ENABLE) & 0x00000001LU)){ /* ←DMA転送中 ならば */
				SetReg(EMMAxTL_DMAC_LIST_RST, 0x00000001LU);             /* DMA転送 停止 */
				while(GetReg(EMMAxTL_DMAC_LIST_ENABLE) & 0x00000001LU){ /* 停止 完了待ち */
					;
				}
			}
			/* DMA転送中でない つまり DMA転送が未使用 or すでに完了ならば _dma0_not_active をセット */
			else{
				_dma0_not_active = 1;
/*	printk("<0>--------> DMA Complete! a\n"); */
			}
		}
		/* DMA転送を再開 */
		else{ /* ↓上位4bitはrsv */   /* [17] next_en↓ */       /*   ただし 転送すべきデータが残っている */
			if((reg->size << 4)|| (reg->config & 0x00020000LU)){ /* ←または 次の 転送LIST があるならば   */
				SetReg(EMMAxTL_DMAC_LIST_ENABLE, 0x00000001LU); /* DMA転送 再開 */
			}
			/* DMA転送がすでに完了ならば _dma0_not_active をセット */
			else{
				_dma0_not_active = 1;
/*	printk("<0>--------> DMA Complete! b\n"); */
			}
		}
	}
}
/*EXPORT_SYMBOL(physmap_flash_pause_dma_by_bootloader); physmap_flash_pause_dma_by_bootloader() は
本drivers/mtd/maps/physmap.c と mtd/chips/cfi_cmdset_0002.c 内でのみ使われていて cfi_cmdset_0002.c と
本physmap.c は １つのモジュールとして必ず一緒にビルドされるので EXPORT_SYMBOL() 不要 */

#endif /* 2009.10.30 Add for Bootloaderで仕掛けたNOR Flash->DDRへのDMA転送 と LinuxのNOR FlashへのWrite 排他 by hiroshi.hasegawa@pixela.co.jp */

static int physmap_flash_remove(struct platform_device *dev)
{
	struct physmap_flash_info *info;
	struct physmap_flash_data *physmap_data;

	info = platform_get_drvdata(dev);
	if (info == NULL)
		return 0;
	platform_set_drvdata(dev, NULL);

	physmap_data = dev->dev.platform_data;

	if (info->mtd != NULL) {
#ifdef CONFIG_MTD_PARTITIONS
		if (info->nr_parts) {
			del_mtd_partitions(info->mtd);
			kfree(info->parts);
		} else if (physmap_data->nr_parts) {
			del_mtd_partitions(info->mtd);
		} else {
			del_mtd_device(info->mtd);
		}
#else
		del_mtd_device(info->mtd);
#endif
		map_destroy(info->mtd);
	}

	if (info->map.virt != NULL)
		iounmap((void *)info->map.virt);

	if (info->res != NULL) {
		release_resource(info->res);
		kfree(info->res);
	}

	return 0;
}

static const char *rom_probe_types[] = { "cfi_probe", "jedec_probe", "map_rom", NULL };
#ifdef CONFIG_MTD_PARTITIONS
static const char *part_probe_types[] = { "cmdlinepart", "RedBoot", NULL };
#endif

static int physmap_flash_probe(struct platform_device *dev)
{
	struct physmap_flash_data *physmap_data;
	struct physmap_flash_info *info;
	const char **probe_type;
	int err;

	physmap_data = dev->dev.platform_data;
	if (physmap_data == NULL)
		return -ENODEV;

       	printk(KERN_NOTICE "physmap platform flash device: %.8llx at %.8llx\n",
	    (unsigned long long)dev->resource->end - dev->resource->start + 1,
	    (unsigned long long)dev->resource->start);

	info = kmalloc(sizeof(struct physmap_flash_info), GFP_KERNEL);
	if (info == NULL) {
		err = -ENOMEM;
		goto err_out;
	}
	memset(info, 0, sizeof(*info));

	platform_set_drvdata(dev, info);

	info->res = request_mem_region(dev->resource->start,
			dev->resource->end - dev->resource->start + 1,
			dev->dev.bus_id);
	if (info->res == NULL) {
		dev_err(&dev->dev, "Could not reserve memory region\n");
		err = -ENOMEM;
		goto err_out;
	}

	info->map.name = dev->dev.bus_id;
	info->map.phys = dev->resource->start;
	info->map.size = dev->resource->end - dev->resource->start + 1;
	info->map.bankwidth = physmap_data->width;
	info->map.set_vpp = physmap_data->set_vpp;

	info->map.virt = ioremap(info->map.phys, info->map.size);
	if (info->map.virt == NULL) {
		dev_err(&dev->dev, "Failed to ioremap flash region\n");
		err = EIO;
		goto err_out;
	}

	simple_map_init(&info->map);

	/* 2009.10.30 Add for Bootloaderで仕掛けたNOR Flash->DDRへのDMA転送 と LinuxのNOR FlashへのWrite 排他 by hiroshi.hasegawa@pixela.co.jp */
	physmap_flash_pause_dma_by_bootloader(1); /* DMA転送を一時停止 (do_map_probe()内で NOR Flashへ Manufacture ID値取得
	                                             コマンドがWriteされる(コマンドがWriteされた直後は DMA転送で NOR Flashから
	                                             Readされるデータに Manufacture ID値が読みだされてしまうから)ため)  */

	probe_type = rom_probe_types;
	for (; info->mtd == NULL && *probe_type != NULL; probe_type++)
		info->mtd = do_map_probe(*probe_type, &info->map);
	if (info->mtd == NULL) {
		dev_err(&dev->dev, "map_probe failed\n");
		err = -ENXIO;
		
		/* 2009.10.30 Add for Bootloaderで仕掛けたNOR Flash->DDRへのDMA転送 と LinuxのNOR FlashへのWrite 排他 by hiroshi.hasegawa@pixela.co.jp */
		physmap_flash_pause_dma_by_bootloader(0); /* ↑での DMA転送の一時停止を解除 */
		
		goto err_out;
	}
	info->mtd->owner = THIS_MODULE;

	/* 2009.10.30 Add for Bootloaderで仕掛けたNOR Flash->DDRへのDMA転送 と LinuxのNOR FlashへのWrite 排他 by hiroshi.hasegawa@pixela.co.jp */
	physmap_flash_pause_dma_by_bootloader(0); /* ↑での DMA転送の一時停止を解除 */

#ifdef CONFIG_MTD_PARTITIONS
	err = parse_mtd_partitions(info->mtd, part_probe_types, &info->parts, 0);
	if (err > 0) {
		add_mtd_partitions(info->mtd, info->parts, err);
		return 0;
	}

	if (physmap_data->nr_parts) {
		printk(KERN_NOTICE "Using physmap partition information\n");
		add_mtd_partitions(info->mtd, physmap_data->parts,
						physmap_data->nr_parts);
		return 0;
	}
#endif

	add_mtd_device(info->mtd);
	return 0;

err_out:
	physmap_flash_remove(dev);
	return err;
}

static struct platform_driver physmap_flash_driver = {
	.probe		= physmap_flash_probe,
	.remove		= physmap_flash_remove,
	.driver		= {
		.name	= "physmap-flash",
	},
};


#ifdef CONFIG_MTD_PHYSMAP_LEN
#if CONFIG_MTD_PHYSMAP_LEN != 0
#warning using PHYSMAP compat code
#define PHYSMAP_COMPAT
#endif
#endif

#ifdef PHYSMAP_COMPAT
static struct physmap_flash_data physmap_flash_data = {
	.width		= CONFIG_MTD_PHYSMAP_BANKWIDTH,
};

static struct resource physmap_flash_resource = {
	.start		= CONFIG_MTD_PHYSMAP_START,
	.end		= CONFIG_MTD_PHYSMAP_START + CONFIG_MTD_PHYSMAP_LEN - 1,
	.flags		= IORESOURCE_MEM,
};

static struct platform_device physmap_flash = {
	.name		= "physmap-flash",
	.id		= 0,
	.dev		= {
		.platform_data	= &physmap_flash_data,
	},
	.num_resources	= 1,
	.resource	= &physmap_flash_resource,
};

void physmap_configure(unsigned long addr, unsigned long size,
		int bankwidth, void (*set_vpp)(struct map_info *, int))
{
	physmap_flash_resource.start = addr;
	physmap_flash_resource.end = addr + size - 1;
	physmap_flash_data.width = bankwidth;
	physmap_flash_data.set_vpp = set_vpp;
}

#ifdef CONFIG_MTD_PARTITIONS
void physmap_set_partitions(struct mtd_partition *parts, int num_parts)
{
	physmap_flash_data.nr_parts = num_parts;
	physmap_flash_data.parts = parts;
}
#endif
#endif

static int __init physmap_init(void)
{
	int err;

	err = platform_driver_register(&physmap_flash_driver);
#ifdef PHYSMAP_COMPAT
	if (err == 0)
		platform_device_register(&physmap_flash);
#endif

	return err;
}

static void __exit physmap_exit(void)
{
#ifdef PHYSMAP_COMPAT
	platform_device_unregister(&physmap_flash);
#endif
	platform_driver_unregister(&physmap_flash_driver);
}

module_init(physmap_init);
module_exit(physmap_exit);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("David Woodhouse <dwmw2@infradead.org>");
MODULE_DESCRIPTION("Generic configurable MTD map driver");
