//#define __KMC_MODULE_DEBUG
//#define __KMC_MODULE_NAME "/emma_driver/emtd_nand_drv.ko"

/*
 * emtd_nand.c
 * NAND mtd ドライバ for EMMA3TL2S
 *
 * H.HASEGAWA
 * 2011.03.28
 *
 * - 新規の MTD NAND ドライバ を 組み込むには MTDが必要とする APIを用意立てして
 *   nand_base.c 内にある nand_scan() へそれらを手渡し＆Callする処理を実装すればよい.
 *
 * - MTDが必要とする API は include/linux/mtd/nand.h 内の構造体変数型 struct nand_chip の
 *   メンバになります. ただし 全メンバのAPIを用意立てする必要はありません.
 *   カスタマイズが必要な API, Pixela内製のEMMA2TL基盤の 環境セッティング データ を
 *   本 emtd_nand.c (ドライバモジュール emtd_nand.ko として) に用意して, モジュールの
 *   初期化関数 emtd_init() 内にて nand_scan() へそれらを 手渡し＆Call すれば
 *   カスタマイズが不要(デフォルトでOK)な API or 環境セッティング は設定されます.
 *
 * - API or 環境セッティング について 置き換えが必要か？ カスタマイズが可能か？ は,
 *   include/linux/mtd/nand.h 内の struct nand_chip についてのコメントを参照のこと.
 *   [BOARDSPECIFIC] : プラットフォーム依存(つまり emtd_init() で設定必須)
 *   [REPLACEABLE]   : 置き換え or カスタマイズ可
 *   [INTERN]        : (基本的に)置き換え不要
 *
 * - NAND Flash の 消去/書き込み/読み込み 手順は trunk/tool/nand_flash_write を参考にしてください.
 *
 * - 本 NAND mtd ドライバ を作成するにあたり ts7250.c を参考にしました.
 *
 * 2010.02.01 H.HASEGAWA	- yaffsが 書き込みエラー時に 空き領域をすべて 不良ブロック化してしまう
 *                      	  不具合を抱えているため 起動時に 各パーティションの"後天性"不良ブロックの割合を
 *                      	  チェックして 不良ブロック数が20%を超えていたら "後天性"不良ブロックを一度 
 *                      	  ブロック消去＆書き込み＆読み込みチェックを行い BBTを消去して MTDにBBTの再構築を
 *                      	  行ってもらう処理を追加
 *
 * 2010.02.08 H.HASEGAWA	- NAND Flashは 各pageの前半256byte中 or 後半256byte中 or spareエリア16byte中 に
 *                      	  1bitのデータ化けが発生するのが当たり前のデバイスで, 前半256byte中 or 後半256byte中
 *                      	  は ECCによって訂正処理がなされているが, spareエリアにおいて1bitデータ化けへの対策が
 *                      	  抜けていたため 対策処理を追加
 *
 * 2011.03.28 H.HASEGAWA	- EMMA3TL2S用へ変更
 *
 * Copyright 2010 PIXELA CORPORATION. All rights reserved.
 */

#include <linux/module.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/err.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/types.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/nand.h>
#include <linux/mtd/nand_ecc.h>
#include <linux/mtd/compatmac.h>
#include <linux/interrupt.h>
#include <linux/bitops.h>
#include <linux/leds.h>
#include <asm/io.h>
#include <asm/emma3tl/emma3tl.h>

#ifdef CONFIG_MTD_PARTITIONS
#include <linux/mtd/partitions.h>
#endif

#define DEBUG_VALID 0 // debug message on/off
#if DEBUG_VALID 
	#define DEBUG_MTD(fmt,args...) printk(KERN_EMERG "%s():" fmt, __FUNCTION__, ##args)
#else 
	#define DEBUG_MTD(fmt,args...)
#endif

/*
 * Ready/BusyピンのReady状態検出を 立ち上がりエッジにて検出する場合に コメントイン
 * (Page Read or Page Write or Block Erase時に Ready/BusyピンがBusy状態になる(立ち下がる)よりも
 *  前に, Ready/Busyピンのレベルを見て, 誤ってReady状態と検出してしまうことを防ぎます)
 */
#define USE_DETECT_RB_LINE_READY_BY_RISING_EDGE

/*
 * 外部モジュール参照関数
 */
/* 2009.09.04 Add for NORとNAND Flashへのアクセス排他 by hiroshi.hasegawa@pixela.co.jp */
extern void cfi_lock_0002(int lock); /* in drivers/mtd/chips/cfi_cmdset_0002.c */

/* 
 * Values specific to the emtd board
 */
#define SZ_1K			(1024)
#define SZ_1M			(SZ_1K*SZ_1K)
#define PAGE_1BLOCK		(32)
#define SZ_1PAGE		(512)
#define SZ_1BLOCK		(SZ_1PAGE*PAGE_1BLOCK)
#define NAND_SIZE64		(64*SZ_1M) 		// Nand Memory size : 64Mbyte(512Mbit)
#define NAND_SIZE32		(32*SZ_1M) 		// Nand Memory size : 32Mbyte(256Mbit)
#define BLOCK_BBT		(4)	// 不良ブロックテーブル : 4block(64Kbyte bbt_main_descr を参照)

/*
 * パーティション数＆サイズ
 */
#define NUM_PARTITIONS	(6) // THでは8
#define BBT_SIZE		(BLOCK_BBT*SZ_1BLOCK)
#define KERNEL_SIZE		(4*SZ_1M)
#define FREE_AREA_SIZE	(32*SZ_1M)
#define FLREC_LOG_SIZE	(5*SZ_1M)
#define UPP_SIZE		(3*SZ_1M)
#define APP_SET_SIZE 	(4*SZ_1M)

/*
 * EMMA3TL2S ROM Interface レジスタ
 */
#define EMMA_ROMIF_REG_BASE_ADR		EMMA_ROMIF_BASE	// EMMA3TL2S は 0xb000a000
#define EMMA_ROMIF_REG_ROM_MODE		(EMMA_ROMIF_REG_BASE_ADR + 0x000)
#define EMMA_ROMIF_REG_ACS_TM		(EMMA_ROMIF_REG_BASE_ADR + 0x004)
#define EMMA_ROMIF_REG_IF_MODE		(EMMA_ROMIF_REG_BASE_ADR + 0x008)
#define EMMA_ROMIF_REG_NAND_CTRL0	(EMMA_ROMIF_REG_BASE_ADR + 0x010)
#define EMMA_ROMIF_REG_NAND_CTRL1L	(EMMA_ROMIF_REG_BASE_ADR + 0x014)
#define EMMA_ROMIF_REG_NAND_CTRL1H	(EMMA_ROMIF_REG_BASE_ADR + 0x018)
#define EMMA_ROMIF_REG_ROM_STATUS	(EMMA_ROMIF_REG_BASE_ADR + 0x020)
#define EMMA_ROMIF_REG_INT_STATUS	(EMMA_ROMIF_REG_BASE_ADR + 0x028)

#define REG_NAND_CTRL0_CE_ENABLE	(0x100) // EMMA3TL2S は NAND FlashのCS は CS0
#define REG_NAND_CTRL0_ADR_CMD_SND	(0x01)
#define REG_NAND_CTRL0_ADR_ENABLE	(0x10)
#define REG_NAND_CTRL0_ADR_NUM(n)	(((n - 1) & 0x7) << 12)
#define REG_NAND_CTRL0_CMD0_ENABLE	(0x04)
#define REG_NAND_CTRL0_CMD0_SHIFT	(16)
#define REG_NAND_CTRL0_CMD1_SHIFT	(24)

/*
 * マクロ関数定義
 */
#define SetReg( adr, dat )			(*(volatile uint32_t *)(EMMA_BASE | (adr)) = dat)
#define GetReg( adr )				(*(volatile uint32_t *)(EMMA_BASE | (adr)))
// NAND Flashへの コマンド,アドレスの発行 完了待ち or 済みチェック
#define WAIT_ADR_CMD_DONE()			do{ ; }while(GetReg(EMMA_ROMIF_REG_NAND_CTRL0) & REG_NAND_CTRL0_ADR_CMD_SND)

/*
 * NAND Flash制御ステータス
 */
#define NAND_STATUS_IDLE			(0x00000000LU)
#define NAND_STATUS_CE_ENABLE		(0x00000001LU)	// CEピンLow(Enable)状態
#ifdef USE_DETECT_RB_LINE_READY_BY_RISING_EDGE
#define NAND_STATUS_WAIT_READY_INT	(0x00000010LU)
#endif
#define NAND_STATUS_ADR_NUM_MSK		(0x0000ff00LU)
#define NAND_STATUS_ADR_NUM(status)	(((uint8_t *)&status)[2])
static uint32_t gl_status = NAND_STATUS_IDLE;

/*
 * 関数プロトタイプ宣言
 */
static void emtd_cmd_ctrl(struct mtd_info *mtd, int dat, unsigned int ctrl);
static int emtd_dev_ready(struct mtd_info *mtd);
static uint8_t emtd_read_byte(struct mtd_info *mtd);
static u16 emtd_read_word(struct mtd_info *mtd);
static void emtd_write_buf(struct mtd_info *mtd, const uint8_t *buf, int len);
static void emtd_read_buf(struct mtd_info *mtd, uint8_t *buf, int len);
static int emtd_verify_buf(struct mtd_info *mtd, const uint8_t *buf, int len);
static int emtd_scan_bbt(struct mtd_info *mtd);
static void _check_bbt_and_recover_for_yaffs_bug(struct mtd_info *mtd);

/*
 * MTD structure for Osbourne board
 */
static struct mtd_info *osbourne_mtd = NULL;

#ifdef CONFIG_MTD_PARTITIONS
static const char *partition_probes[] = { "cmdlinepart", NULL };

/*
 * Define static partitions for flash device
 */
static struct mtd_partition osbourne_nand_partition_info64[]= {
	/* 64Mbyte(512Mbyte) NAND Flash用 パーティション設定 */
	{ // mtd8
		.name = "OSBOURNE : Free Area",
		.offset = BBT_SIZE,
		.size = FREE_AREA_SIZE
	},
	{ // mtd9
		.name = "OSBOURNE : Kernel(vmlinux)",
		.offset = MTDPART_OFS_APPEND,
		.size = KERNEL_SIZE
	},
	{ // mtd10
		.name = "OSBOURNE : Flight Record Log(YAFFS)",
		.offset = MTDPART_OFS_APPEND,
		.size = FLREC_LOG_SIZE
	},
	{ // mtd11
		.name = "OSBOURNE : Updater(YAFFS)",
		.offset = MTDPART_OFS_APPEND,
		.size = UPP_SIZE
	},
	{ // mtd12
		.name = "OSBOURNE : Application Setting data(YAFFS)",
		.offset = MTDPART_OFS_APPEND,
		.size = APP_SET_SIZE
	},
	{ // mtd13
		.name = "OSBOURNE : MassStorage(YAFFS)",
		.offset = MTDPART_OFS_APPEND,
		.size = NAND_SIZE64 - (BBT_SIZE + KERNEL_SIZE + FREE_AREA_SIZE + FLREC_LOG_SIZE + UPP_SIZE + APP_SET_SIZE)
	}
};
static struct mtd_partition osbourne_nand_partition_info32[]= {
	/* 32Mbyte(256Mbit) NAND Flash用 パーティション設定 */
	/* (mtd番号を64Mbyteのそれと変えないように Dummyパーティションを用意する) */
	{ // mtd8
		.name = "OSBOURNE : Dummy partition",
		.offset = 0,
		.size = SZ_1BLOCK
	},
	{ // mtd9
		.name = "OSBOURNE : Kernel(vmlinux)",
		.offset = BBT_SIZE,
		.size = KERNEL_SIZE
	},
	{ // mtd10
		.name = "OSBOURNE : Flight Record Log(YAFFS)",
		.offset = MTDPART_OFS_APPEND,
		.size = FLREC_LOG_SIZE
	},
	{ // mtd11
		.name = "OSBOURNE : Updater(YAFFS)",
		.offset = MTDPART_OFS_APPEND,
		.size = UPP_SIZE
	},
	{ // mtd12
		.name = "OSBOURNE : Application Setting data(YAFFS)",
		.offset = MTDPART_OFS_APPEND,
		.size = APP_SET_SIZE
	},
	{ // mtd13
		.name = "OSBOURNE : MassStorage(YAFFS)",
		.offset = MTDPART_OFS_APPEND,
		.size = NAND_SIZE32 - (BBT_SIZE + KERNEL_SIZE + FLREC_LOG_SIZE + UPP_SIZE + APP_SET_SIZE)
	}
};
#endif /* CONFIG_MTD_PARTITIONS */

/*
 * ECCデータのレイアウト
 * 
 * EMMA2THのECCレイアウトへカスタマイズする
 * (Defaultは nand_base.c line 65 ～ )
 * 
 * Defaultは スペアエリア(Page終端の16byte)の 0 ～ 3, 6 ～ 7byte だが
 * 8 ～ 10, 13 ～ 15byte (YAFFSのOOBデータ内のECCレイアウトと同じ) へ変更
 */
static struct nand_ecclayout emtd_oob_16 = {
	.eccbytes = 6,
	.eccpos = {8, 9, 10, 13, 14, 15}, /* ←各Page終端16byteの冗長部部に 左記のOffsetにてECC値を書き込む */
	.oobfree = {
		{.offset = 0,
		 . length = 4},
		{.offset = 6,
		 . length = 2},
		{.offset = 11,
		 . length = 2}}
};

/*
 * 不良ブロックテーブルのデータレイアウト
 * 
 * 下記の理由により Defaultのものでなく カスタマイズする
 * (Defaultは nand_bbt.c line 1117 ～ )
 * 
 * a.不良ブロックテーブル を NAND Flashの先頭ブロックに格納したいから
 *   (Defaultは NAND Flashの最終端にある4つの良ブロックを使用している)
 * b.ECCデータ格納位置 を YAFFSを利用するために変更したが マジックパターンデータ と
 *   バージョン値 の格納位置 が 被ってしまっているため
 *   (Defaultは スペアエリア(Page終端の16byte)の
 *    マジックパターンデータ : 8～11byte, バージョン値 : 12byte を使用しているが
 *    ECC格納位置が 8～15byte であり, 被ってしまっています)
 */

static uint8_t bbt_pattern[] = {'b'}; /* Default {'B', 'b', 't', '0' } の 4文字だが 1文字で十分 */
static uint8_t mirror_pattern[] = {'m'}; /* Default {'1', 't', 'b', 'B' } の 4文字だが 1文字で十分 */

static struct nand_bbt_descr bbt_main_descr = {
	// a. の理由で NAND_BBT_LASTBLOCK をコメントアウト
	.options = /* NAND_BBT_LASTBLOCK | */ NAND_BBT_CREATE | NAND_BBT_WRITE
		| NAND_BBT_2BIT | NAND_BBT_VERSION | NAND_BBT_PERCHIP,
	// b. の理由で 下記のように変更
	.offs =	3,		/* Default  8 (ECCにも不良ブロックMark(5byte)と被らない3byteへ)*/
	.len = 1,		/* Default  4 (1文字へ変更したので 1 へ)*/
	.veroffs = 7,	/* Default 12 (ECCと被らない7byteへ)*/
	.maxblocks = BLOCK_BBT, /* 容量は Default 4block でＯＫ*/
	.pattern = bbt_pattern /* 4文字も要らない, 1文字でよい */
};

static struct nand_bbt_descr bbt_mirror_descr = {
	// a. の理由で NAND_BBT_LASTBLOCK をコメントアウト
	.options = /* NAND_BBT_LASTBLOCK | */ NAND_BBT_CREATE | NAND_BBT_WRITE
		| NAND_BBT_2BIT | NAND_BBT_VERSION | NAND_BBT_PERCHIP,
	// b. の理由で 下記のように変更
	.offs =	3,		/* Default  8 (ECCにも不良ブロックMark(5byte)と被らない3byteへ)*/
	.len = 1,		/* Default  4 (1文字へ変更したので 1 へ)*/
	.veroffs = 7,	/* Default 12 (ECCと被らない7byteへ)*/
	.maxblocks = BLOCK_BBT, /* 容量は Default 4block でＯＫ*/
	.pattern = mirror_pattern /* 4文字も要らない, 1文字でよい */
};

/*
 * 不良ブロックマーキング位置
 * 
 * nand_bbt.c line 1094 ～ をコピーしただけで, Defaultのまま
 * (各ブロック先頭2pageのspareエリア16byteの どちらかの spare[5] が
 *  0xff と 7bit以上一致していれば 良, そうでなければ 不良ブロックとなる.
 *  ※NAND Flashは 256byte中 or OOB中 に 1bitのデータ化けが発生するのが
 *    当たり前なので 7bit以上一致していれば 良 となる)
 */

static uint8_t scan_ff_pattern[] = { 0xff, 0xff };

static struct nand_bbt_descr smallpage_flashbased = {
	.options = NAND_BBT_SCAN2NDPAGE,
	.offs = 5,
	.len = 1,
	.pattern = scan_ff_pattern
};

/*
 * Main initialization routine
 */
static int __init emtd_init(void)
{
	struct nand_chip *chip;
	const char *part_type = 0;
	int mtd_parts_nb = 0;
	struct mtd_partition *mtd_parts = 0;
	
	DEBUG_MTD(" NAND Flash ROM driver for PIXELA board\n");
	
	// Allocate memory for MTD device structure and private data
	osbourne_mtd = kmalloc(sizeof(struct mtd_info) + sizeof(struct nand_chip), GFP_KERNEL);
	if(!osbourne_mtd){
		printk(KERN_ERR "err : emtd_init() kmalloc failed\n");
		return -ENOMEM;
	}
	
	// Get pointer to private data
	chip = (struct nand_chip *)(&osbourne_mtd[1]);
	
	// Initialize structures 
	memset((char *)osbourne_mtd, 0, sizeof(struct mtd_info));
	memset((char *)chip, 0, sizeof(struct nand_chip));
	
	// Link the private data with the MTD structure
	osbourne_mtd->priv = chip;
	osbourne_mtd->owner = THIS_MODULE;
	
	// insert callbacks 1
	// まず struct nand_chip のメンバーの内 プラットフォーム依存の [BOARDSPECIFIC] メンバーを設定
	chip->IO_ADDR_R = 
	chip->IO_ADDR_W = (void *)(EMMA_BASE | EMMA_NAND_BASE); // EMMA3TL2S は NAND Flash のメモリ空間が 0xbc000000～
	chip->cmd_ctrl = emtd_cmd_ctrl;
	chip->dev_ready = emtd_dev_ready;
	chip->ecc.mode = NAND_ECC_SOFT;	
	chip->ecc.layout = &emtd_oob_16; // (←YAFFS用にECC格納位置をカスタマイズ)
	chip->options = NAND_NO_PADDING | NAND_COPYBACK |
					NAND_USE_FLASH_BBT; // (←不良ブロックテーブルを NAND Flash上に記録するため)
	chip->chip_delay = 15; // 15 us command delay time
	
	// insert callbacks 2
	// 次に struct nand_chip のメンバーの内 置き換えが必要な [REPLACEABLE] メンバーを設定
	// (｜NAND Flashへのアクセスを MMU経由のiomap経由でなく ダイレクトアクセスさせるため.
	//  ↓PCIバスやグラフィックに使用するわけでない(リニアなアドレスは不要)なので)
	chip->read_byte = emtd_read_byte;
	chip->read_word = emtd_read_word;
	chip->write_buf = emtd_write_buf;
	chip->read_buf = emtd_read_buf;
	chip->verify_buf = emtd_verify_buf;
	// (↓不良ブロックテーブルをカスタマイズするため)
	chip->bbt_td = &bbt_main_descr;
	chip->bbt_md = &bbt_mirror_descr;
	
	// insert callbacks 3
	// 最後に yaffsが 書き込みエラー時に 空き領域をすべて 不良ブロック化してしまう
	// 不具合を抱えているため 起動時に 各パーティションの"後天性"不良ブロックの割合を
	// チェックして 不良ブロック数が20%を超えていたら "後天性"不良ブロックを一度 
	// ブロック消去＆書き込み＆読み込みチェックを行い BBTを消去して MTDにBBTの再構築を
	// 行ってもらう
	// その処理を仕込んだ scan_bbt() に差し替えます
	chip->scan_bbt = emtd_scan_bbt;
	
	// Scan to find existence of the device
	if(nand_scan(osbourne_mtd, 1)){
		printk(KERN_ERR "err : emtd_init() nand_scan failed\n");
		kfree(osbourne_mtd);
		return -ENXIO;
	}
	
#ifdef CONFIG_MTD_PARTITIONS
	osbourne_mtd->name = "osbourne_nand";
	mtd_parts_nb = parse_mtd_partitions(osbourne_mtd, partition_probes, &mtd_parts, 0);
	if(mtd_parts_nb > 0){
		part_type = "command line";
	}
	else {
		mtd_parts_nb = 0;
	}
#endif
	if(mtd_parts_nb == 0){
		if(chip->chipsize < NAND_SIZE64){
			mtd_parts = osbourne_nand_partition_info32; /* 32Mbyte NAND Flash */
		}
		else{
			mtd_parts = osbourne_nand_partition_info64; /* 64Mbyte NAND Flash */
		}
		mtd_parts_nb = NUM_PARTITIONS;
		part_type = "static";
	}
	
	// Register the partitions
	DEBUG_MTD("Using %s partition definition\n", part_type);
	add_mtd_partitions(osbourne_mtd, mtd_parts, mtd_parts_nb);
	
	// Return happy
	return 0;
}

/*
 * Clean up routine
 */
static void __exit emtd_cleanup(void)
{
	// Unregister the device
	del_mtd_device(osbourne_mtd);
	
	// Free the MTD device structure 
	kfree(osbourne_mtd);
	
	SetReg(EMMA_ROMIF_REG_NAND_CTRL0, 0);
}

module_init(emtd_init);
module_exit(emtd_cleanup);

/**
 * emtd_cmd_ctrl - [BOARDSPECIFIC] hardwarespecific funtion for controlling ALE/CLE/nCE.
 * @mtd:	MTD device structure
 * @dat:	NAND Flash へのコマンド
 * @ctrl:	ALE/CLE/nCEピン制御値
 *
 * Also used to write command and address
 */
static void emtd_cmd_ctrl(struct mtd_info *mtd, int dat, unsigned int ctrl)
{
	unsigned int cur_ctrl; // 現在のピン制御値
	uint32_t adr_num;
	static unsigned int gl_ctrl = 0; // 前回のif(ctrl & NAND_CTRL_CHANGE)にてCall時のピン制御値
	static uint32_t gl_adrs[2] = {0}; // 設定されているアドレス値
#ifdef USE_DETECT_RB_LINE_READY_BY_RISING_EDGE
	static int gl_dat[] = {NAND_CMD_NONE, NAND_CMD_NONE}; // 過去2つのコマンド履歴
	#define NAND_CMD_ADR	(NAND_CMD_NONE - 1)
#endif
	
	if(ctrl & NAND_CTRL_CHANGE){
		// [フェールセーフ] ありえないが ALE と CLE の両方が有効な場合
		if((ctrl & (NAND_ALE | NAND_CLE)) == (NAND_ALE | NAND_CLE)){
			printk(KERN_ERR "err : emtd_cmd_ctrl() CLE and ALE settled both !!!\n");
			if(gl_ctrl & NAND_ALE){ // 後勝ちで ALE or CLE どちらかにしておく
				cur_ctrl = ctrl & ~NAND_ALE;
			}
			else{
				cur_ctrl = ctrl & ~NAND_CLE;
			}
		}
		else{
			cur_ctrl = ctrl;
		}
		
		// CEピン(チップセレクト)が 無効 -> 有効 へ変化
	 	if(!(gl_ctrl & NAND_NCE) && (cur_ctrl & NAND_NCE)){
	 		// NAND Flash制御ステータス が IDLE かチェック
			if(gl_status){ // ←ココでIDLEでなければ おかしい！
				printk(KERN_ERR "err : emtd_cmd_ctrl() Exist not executed commands or addresses gl_status=0x%08X a\n", gl_status);
			}
			
			if(!(gl_status & NAND_STATUS_CE_ENABLE)){
				/* 2009.09.04 Add for NORとNAND Flashへのアクセス排他 by hiroshi.hasegawa@pixela.co.jp */
				cfi_lock_0002(1); /* NOR Flashアクセスをロック(NANDへのCEピンをEnableにするので) */
				
				/* 2011.03.28 Add for EMMA3TL2S by hiroshi.hasegawa@pixela.co.jp */
				/* EMMA2TLでは NOR と NAND Flash が同じデータバスに接続されており
				   NOR と NAND それぞれにCS(Chip Selectピン)0 と CS1 が割り当てられ
				   CS1をアサインすることで NAND Flash へアクセス出来ていた.
				   しかし EMMA3TL2Sでは NOR と NAND Flash が別々のデータバスに接続
				   されており CSは共に CS0 が割り当てられているため NAND へアクセス
				   するには EMMA_BHIF_STRAP_0 の NOR_USE を 0b へ変更しなければなりません */
				// NOR_USE を 0b(NOR から NAND Flash へのアクセスに切り替え) へ
				SetReg(EMMA_BHIF_STRAP_0_W, GetReg(EMMA_BHIF_STRAP_0_W) & (~0x00000080));
				// ROM_MODE0 を 10b(NAND Flash ROM) へ
				SetReg(EMMA_ROMIF_REG_ROM_MODE, (GetReg(EMMA_ROMIF_REG_ROM_MODE) & (~0x00030000)) | 0x00020000);
				
				SetReg(EMMA_ROMIF_REG_NAND_CTRL0, REG_NAND_CTRL0_CE_ENABLE);
				gl_status |= NAND_STATUS_CE_ENABLE;
			}
			
#ifdef USE_DETECT_RB_LINE_READY_BY_RISING_EDGE
			// コマンド履歴をクリア
			gl_dat[0] = NAND_CMD_NONE;
#endif
		}
		
		// ALE(アドレスラッチ)ピンが 有効 -> 無効 へ変化 ならば アドレス発行
	 	if(!(cur_ctrl & NAND_ALE) && (gl_ctrl & NAND_ALE)){
	 		adr_num = NAND_STATUS_ADR_NUM(gl_status);
	 		if(adr_num){
				// コマンド,アドレスの発行 完了待ち or 済みチェック
				WAIT_ADR_CMD_DONE();
	 			
	 			// 発行するアドレスをレジスタへセット
	 			SetReg(EMMA_ROMIF_REG_NAND_CTRL1L, gl_adrs[1]);
	 			if(4 < adr_num){
	 				SetReg(EMMA_ROMIF_REG_NAND_CTRL1H, gl_adrs[0]);
	 			}
	 			
#ifdef USE_DETECT_RB_LINE_READY_BY_RISING_EDGE
				// Page Read の場合は Ready/Busyピン検出を 立ち上がりエッジ検出 へ
				switch(gl_dat[0]){
				case NAND_CMD_READ0:
				case NAND_CMD_READ1:
				case NAND_CMD_READOOB:
					SetReg(EMMA_ROMIF_REG_INT_STATUS, 0); // 立ち上がりエッジ検出ビット を クリア
					gl_status |= NAND_STATUS_WAIT_READY_INT; // 立ち上がりエッジ検出フラグ を セット
					break;
				default:
					break;
				}
#endif
				
				// アドレス発行
				SetReg(EMMA_ROMIF_REG_NAND_CTRL0, REG_NAND_CTRL0_CE_ENABLE |
				       REG_NAND_CTRL0_ADR_ENABLE | REG_NAND_CTRL0_ADR_NUM(adr_num) | REG_NAND_CTRL0_ADR_CMD_SND);
				NAND_STATUS_ADR_NUM(gl_status) = 0; // アドレス設定数をクリア
				
#ifdef USE_DETECT_RB_LINE_READY_BY_RISING_EDGE
				// コマンド履歴を追加
				gl_dat[1] = gl_dat[0]; gl_dat[0] = NAND_CMD_ADR;
#endif
	 		}
	 		else{
	 			printk(KERN_ERR "err : emtd_cmd_ctrl() Nothing address dat=0x%02X, ctrl=0x%02X\n", dat, ctrl);
	 		}
		}
	}
	else{
		cur_ctrl = gl_ctrl & ~NAND_CTRL_CHANGE;
	}
	
	// アドレス設定 or コマンド発行
	if(dat != NAND_CMD_NONE){
		// ALE(アドレスラッチ)ピンが有効ならば アドレス設定
		// (発行はまだせず, ピン制御値が ALE(アドレスラッチ)ピンが 有効 -> 無効 へ変化したら 発行する.
		//  なぜなら, 1byte毎に発行すると, ALEピンがアドレス1byte発行毎に変化して, NAND Flashがアドレス
		//  全アドレス発行の終了を把握出来ないため)
		if(cur_ctrl & NAND_ALE){
			adr_num = NAND_STATUS_ADR_NUM(gl_status);
			if(adr_num < 8){
				((uint8_t *)gl_adrs)[7 - adr_num] = dat & 0xff; // ←※Big Endianなので こうなります
				NAND_STATUS_ADR_NUM(gl_status) = adr_num + 1; // アドレス設定数を更新
			}
			else{
	 			printk(KERN_ERR "err : emtd_cmd_ctrl() Over address dat=0x%02X, ctrl=0x%02X\n", dat, ctrl);
			}
		}
		// CLE(コマンドラッチ)ピンが有効ならば コマンドをセット
		else if(cur_ctrl & NAND_CLE){
			// コマンド,アドレスの発行 完了待ち or 済みチェック
			WAIT_ADR_CMD_DONE();
			
			switch(dat){
#ifdef USE_DETECT_RB_LINE_READY_BY_RISING_EDGE
			// Page Write or Block Erase の場合は Ready/Busyピン検出を 立ち上がりエッジ検出 へ
			case NAND_CMD_PAGEPROG:
				if(gl_dat[1] != NAND_CMD_SEQIN || gl_dat[0] != NAND_CMD_ADR){
					break;
				}
			case NAND_CMD_ERASE2:
				if(gl_dat[1] != NAND_CMD_ERASE1 || gl_dat[0] != NAND_CMD_ADR){
					break;
				}
				SetReg(EMMA_ROMIF_REG_INT_STATUS, 0); // 立ち上がりエッジ検出ビット を クリア
				gl_status |= NAND_STATUS_WAIT_READY_INT; // 立ち上がりエッジ検出フラグ を セット
				break;
#endif
			default:
				break;
			}
			
			// コマンド発行
			SetReg(EMMA_ROMIF_REG_NAND_CTRL0, REG_NAND_CTRL0_CE_ENABLE |
			       ((dat & 0xff) << REG_NAND_CTRL0_CMD0_SHIFT) | REG_NAND_CTRL0_CMD0_ENABLE |
			       REG_NAND_CTRL0_ADR_CMD_SND);
			
#ifdef USE_DETECT_RB_LINE_READY_BY_RISING_EDGE
			// コマンド履歴を追加
			gl_dat[1] = gl_dat[0]; gl_dat[0] = dat;
#endif
		}
		// [フェールセーフ] ありえないが ALE も CLE も無効な状態ならば 単なる1byte書き込みをしておく
		else{
			// コマンド,アドレスの発行 完了待ち or 済みチェック
			WAIT_ADR_CMD_DONE();
			
			printk(KERN_ERR "err : emtd_cmd_ctrl() Set command or address at unknown status dat=0x%02X, ctrl=0x%02X\n", dat, ctrl);
			*((volatile uint8_t *)((struct nand_chip *)mtd->priv)->IO_ADDR_W) = dat & 0xff;
		}
	}
	
	if(cur_ctrl & NAND_CTRL_CHANGE){
		// CEピン(チップセレクト)が 有効 -> 無効 へ変化
		if(!(cur_ctrl & NAND_NCE) && (gl_ctrl & NAND_NCE)){
			// コマンド,アドレスの発行 完了待ち or 済みチェック
			WAIT_ADR_CMD_DONE();
			
			if(gl_status & NAND_STATUS_CE_ENABLE){
				SetReg(EMMA_ROMIF_REG_NAND_CTRL0, 0);
				gl_status &= ~NAND_STATUS_CE_ENABLE;
				
				/* 2011.03.28 Add for NAND Flashへのアクセスを無効へ by hiroshi.hasegawa@pixela.co.jp */
				// NOR_USE を 1b(NAND から NOR Flash へのアクセスに切り替え) へ
				SetReg(EMMA_BHIF_STRAP_0_W, GetReg(EMMA_BHIF_STRAP_0_W) | 0x00000080);
				// ROM_MODE0 を 00b(NOR Flash ROM) へ
				SetReg(EMMA_ROMIF_REG_ROM_MODE, GetReg(EMMA_ROMIF_REG_ROM_MODE) & (~0x00030000));
				
				/* 2009.09.04 Add for NORとNAND Flashへのアクセス排他 by hiroshi.hasegawa@pixela.co.jp */
				cfi_lock_0002(0); /* NOR Flashアクセスをアンロック(NANDへのCEピンをDisableにしたので) */
			}
			
			// NAND Flash制御ステータス が IDLE かチェック
			if(gl_status){ // ←ココでIDLEでなければ おかしい！
				printk(KERN_ERR "err : emtd_cmd_ctrl() Exist not executed commands or addresses gl_status=0x%08X b\n", gl_status);
				gl_status = NAND_STATUS_IDLE;
			}
		}
		
		// ピン制御値を保存
		gl_ctrl = cur_ctrl;
	}
}

/**
 * emtd_dev_ready - [BOARDSPECIFIC] hardwarespecific function for accesing device ready/busy line
 * @mtd:	MTD device structure
 *
 * NAND Flash の Ready/Busyピンの状態を返す
 */
static int emtd_dev_ready(struct mtd_info *mtd)
{
	// コマンド,アドレスの発行 完了待ち or 済みチェック
	WAIT_ADR_CMD_DONE();
	
#ifdef USE_DETECT_RB_LINE_READY_BY_RISING_EDGE
	// Ready/Busyピンの立ち上がりエッジ検出 が有効になっているならば
	if(gl_status & NAND_STATUS_WAIT_READY_INT){
		// Ready/Busyピンの立ち上がりエッジ検出ビット を チェック
		if(GetReg(EMMA_ROMIF_REG_INT_STATUS) & 0x1){
			SetReg(EMMA_ROMIF_REG_INT_STATUS, 0); // 立ち上がりエッジ検出ビット を クリア
			gl_status &= ~NAND_STATUS_WAIT_READY_INT; // 立ち上がりエッジ検出フラグ を クリア
		}
		else{
			return 0;
		}
	}
	// そうでなければ Ready/Busyピンのレベルを返す
#endif
	
	// Ready/Busyピンのレベルを返す
	return (GetReg(EMMA_ROMIF_REG_ROM_STATUS) & 0x1); // 0:Busy/1:Ready
}

/**
 * emtd_read_byte - [REPLACEABLE] read one byte from the chip
 * @mtd:	MTD device structure
 *
 * Default read function for 8bit buswith
 *
 * nand_base.c 内にある nand_read_byte() の置き換え
 */
static uint8_t emtd_read_byte(struct mtd_info *mtd)
{
	// コマンド,アドレスの発行 完了待ち or 済みチェック
	WAIT_ADR_CMD_DONE();
	
	return (*((volatile uint8_t *)((struct nand_chip *)mtd->priv)->IO_ADDR_R));
}

/**
 * emtd_read_word - [REPLACEABLE] read one word from the chip
 * @mtd:	MTD device structure
 *
 * Default read function for 16bit buswith without
 * endianess conversion
 *
 * nand_base.c 内にある nand_read_word() の置き換え
 */
static u16 emtd_read_word(struct mtd_info *mtd)
{
	// コマンド,アドレスの発行 完了待ち or 済みチェック
	WAIT_ADR_CMD_DONE();
	
	return (*((volatile u16 *)((struct nand_chip *)mtd->priv)->IO_ADDR_R));
}

/**
 * emtd_write_buf - [REPLACEABLE] write data from the buffer to the chip
 * @mtd:	MTD device structure
 * @buf:	data buffer
 * @len:	number of bytes to write
 *
 * Default write function for 8bit buswith
 *
 * nand_base.c 内にある nand_write_buf() の置き換え
 */
static void emtd_write_buf(struct mtd_info *mtd, const uint8_t *buf, int len)
{
	uint8_t *r_adr = (uint8_t *)buf;
	uint8_t *r_adr_end = (uint8_t *)&buf[len];
	volatile uint8_t *w_adr = (volatile uint8_t *)((struct nand_chip *)mtd->priv)->IO_ADDR_W;
	
	// コマンド,アドレスの発行 完了待ち or 済みチェック
	WAIT_ADR_CMD_DONE();
	
	while(r_adr < r_adr_end){
		*w_adr = *r_adr;
		r_adr++;
	}
}

/**
 * emtd_read_buf - [REPLACEABLE] read data from the chip into the buffer
 * @mtd:	MTD device structure
 * @buf:	buffer to store date
 * @len:	number of bytes to read
 *
 * Default read function for 8bit buswith
 *
 * nand_base.c 内にある nand_read_buf() の置き換え
 */
static void emtd_read_buf(struct mtd_info *mtd, uint8_t *buf, int len)
{
	uint8_t *w_adr = (uint8_t *)buf;
	uint8_t *w_adr_end = (uint8_t *)&buf[len];
	volatile uint8_t *r_adr = (volatile uint8_t *)((struct nand_chip *)mtd->priv)->IO_ADDR_R;
	
	// コマンド,アドレスの発行 完了待ち or 済みチェック
	WAIT_ADR_CMD_DONE();
	
	while(w_adr < w_adr_end){
		*w_adr = *r_adr;
		w_adr++;
	}
}

/**
 * emtd_verify_buf - [REPLACEABLE] verify buffer contents against the chip data
 * @mtd:	MTD device structure
 * @buf:	buffer containing the data to compare
 * @len:	number of bytes to compare
 *
 * Default verify function for 8bit buswith
 *
 * nand_base.c 内にある nand_verify_buf() の置き換え
 */
static int emtd_verify_buf(struct mtd_info *mtd, const uint8_t *buf, int len)
{
	uint8_t *buf_adr = (uint8_t *)buf;
	uint8_t *buf_adr_end = (uint8_t *)&buf[len];
	volatile uint8_t *r_adr = (volatile uint8_t *)((struct nand_chip *)mtd->priv)->IO_ADDR_R;
	
	// コマンド,アドレスの発行 完了待ち or 済みチェック
	WAIT_ADR_CMD_DONE();
	
	while(buf_adr < buf_adr_end){
		// ｜NOR も NAND Flashでも データ書き込みは まずBlockEraseにてブロック全体を0xffとし
		// ｜データ書き込み時に 値が 0 のbitが 0 に変化することで書き込む(Flashデバイスの基本)
		// ｜この特性を用いて yaffs では BlockEraseしてデータ書き込みを行うのではなく すでに
		// ｜データを書き込み済みのpage上の値が 1 の特定bitを 0 に書きかえることで そのpageの
		// ｜データを無効とする書き込み処理がある.
		// ｜その際の書き込みデータは 0 にしたいbitのみが 0 で その他のbitは 1 という 書き込み
		// ｜データ と 書き込み後のNAND Flash上のデータが異なる 書き込みデータが手渡される.
		// ｜その際に 目的通りの書き込みが出来ているのに Verifyエラーとなってはならないので
		// ｜書き込みデータ と 書き込み後のNAND Flash上のデータ の比較は『書き込みデータの
		// ｜値が 0 のbitが 書き込み後のNAND Flash上のデータにおいて 0 となっているか？』のみ
		// ↓チェックする.
		if((~(*buf_adr)) & (*r_adr)){
			return -EFAULT;
		}
		buf_adr++;
	}
	
	return 0;
}

/**
 * emtd_scan_bbt - [REPLACEABLE] function to scan bad block table
 * @mtd:	MTD device structure
 *
 * This function selects the default bad block table
 * support for the device and calls the nand_scan_bbt function
 *
 * nand_bbt.c 内にある nand_default_bbt() の置き換え
 */
static int emtd_scan_bbt(struct mtd_info *mtd)
{
	struct nand_chip *this = mtd->priv;
	
	/* flash based bad block table is requested */
	if (!this->badblock_pattern) {
		this->badblock_pattern = &smallpage_flashbased;
	}
	
	// yaffsの不具合による不良ブロック化が発生していないかチェックを行い
	// 不良ブロック化が発生している場合には 復帰処理を施す
	_check_bbt_and_recover_for_yaffs_bug(mtd);
	
	return nand_scan_bbt(mtd, this->badblock_pattern);
}

/**
 * int _memcmp_igonore_1bit(uint8_t *a, uint8_t *b, int len)
 * 
 * a と b の値が一致しているかチェックする.
 * ただし 1bitだけの違いは 無視する
 *
 * ※NAND Flashは 各pageの前半256byte中 or 後半256byte中
 *   or spareエリア16byte中 に 1bitのデータ化けが発生するのが
 *   当たり前のデバイスなので, 1bitのデータ化けを無視するために
 *   本関数を利用
 */
static int _memcmp_igonore_1bit(uint8_t *a, uint8_t *b, int len)
{
	int cnt = 0;
	uint8_t tmp, mask;
	
	while(len){
		// 全bitが一致していないならば 1bitづつチェック
		if((*a) ^ (*b)){
			// 不一致bitが 何bitあるか？チェック
			tmp = (*a) ^ (*b);
			mask = 0x01;
			while(mask){
				if(tmp & mask){
					cnt++;
					// 不一致bitが2bit以上見つかったら データ不一致なので 1 を返す
					if(2 <= cnt){
						return 1;
					}
				}
				mask <<= 1;
			}
		}
		
		// 次のbyteへ
		a++; b++;
		len--;
	}
	
	// 不一致bitが1bitだけ or 値が一致しているので 0 を返す
	return 0;
}

/**
 * void _check_bbt_and_recover_for_yaffs_bug(struct mtd_info *mtd)
 * 
 * NAND Flash先頭のBBT(不良ブロック管理テーブル)から 各パーティションの
 * 不良ブロックの割合をチェックして 不良ブロックが20%を超えていたら異常なので
 * "後天性"不良ブロックを一度 ブロック消去＆書き込み＆読み込みチェックを行い
 * BBTを消去して MTDにBBTの再構築を行ってもらう
 */
static void _check_bbt_and_recover_for_yaffs_bug(struct mtd_info *mtd)
{
	uint32_t tmp, mask;
	int block, page, next_ptt_block, cnt, bbt_page_num;
	uint32_t *bbt_buf;
	struct nand_chip *chip = mtd->priv;
	struct nand_bbt_descr *bd = chip->badblock_pattern;
	struct nand_bbt_descr *td;
	struct mtd_partition *mtd_parts;
	uint8_t *oob_data;
	int bad_ptt_num;
	
	// NAND Flashが無い or 容量が0byte ならば(ありえないはずですが)
	if(chip->numchips == 0 || chip->chipsize == 0){
		printk(KERN_ERR "err : Nothing NAND Flash [%d,%lu]\n", chip->numchips, chip->chipsize);
		return;
	}
	
	// BBTデータサイズ(page単位)を算出
	//  各blockの状態が2bit値で表わされているので  (1pageに満たない↓余剰があれば1page加算されるように加算！)
	//  [BBTデータのpage数] = (2bit * [NAND Flashの全block数] + ([1pageのbit数] - 1)) / 1pageのbit数 
	bbt_page_num = (2L * (chip->chipsize / SZ_1BLOCK) + (8L * SZ_1PAGE - 1)) / (8L * SZ_1PAGE);
	
	// BBT読み込みバッファを確保
	bbt_buf = (uint32_t *)kmalloc(bbt_page_num * SZ_1PAGE, GFP_KERNEL);
	if(!bbt_buf){
		printk(KERN_ERR "err : Can't malloc for NAND Flash BBT [%d]\n", bbt_page_num * SZ_1PAGE);
		return;
	}
	
	// NAND Flash を Activate
	chip->select_chip(mtd, 0);
	
	// BBT読み込み
	td = chip->bbt_td; // BBTデスクリプタをセット(BBTのマジック値等の bbt_main_descr です)
	block = 0;
	while(1){
		// BBTを読み込みつつ NAND Flashの全block分のBBTが書かれているか？＆不良ブロックでないか？ チェック
		cnt = 0; // BBTが書き込まれているpage数
		for(page = 0; page < ((bbt_page_num < 2) ? 2 : bbt_page_num); page++){
			if(page < bbt_page_num){
				// BBTを1pageづつ読み込み
				// ※NAND Flashは 256byte中に1bit のデータ化けが発生するのが 当たり前のデバイスであり
				//   必ず ECCチェック＆訂正 を行って データの読み出し(ecc.read_page()にて)を行うこと！
				tmp = mtd->ecc_stats.failed;
				chip->oob_poi = chip->buffers.oobrbuf; // ←忘れないように！
				chip->cmdfunc(mtd, NAND_CMD_READ0, 0, block * PAGE_1BLOCK + page);
				chip->ecc.read_page(mtd, chip, (uint8_t *)&bbt_buf[page * SZ_1PAGE / sizeof(uint32_t)]);
				// ECCエラー発生＆訂正不可 ならば 抜ける
				if(mtd->ecc_stats.failed != tmp){
					printk(KERN_ERR "err : NAND Flash BBT is Broken [%u]page\n", block*PAGE_1BLOCK + page);
					break;
				}
			}
			else{
				// BBTを読み込み終えたならば OOB領域のみを読み込み
				chip->oob_poi = chip->buffers.oobrbuf; // ←忘れないように！
				chip->ecc.read_oob(mtd, chip, block*PAGE_1BLOCK + page, 1);
			}
			oob_data = chip->oob_poi;
			
			// 不良ブロックMarkあり(つまり 不良ブロック)ならば 抜ける
			if(page < 2 && _memcmp_igonore_1bit(&oob_data[bd->offs], bd->pattern, bd->len) != 0){
				break;
			}
			
			// bbt_pattern[]が書き込まれている(つまり BBTが書き込まれている)ならば
			if(page < bbt_page_num && _memcmp_igonore_1bit(&oob_data[td->offs], td->pattern, td->len) == 0){
				cnt++; // BBTが書き込まれているpage数を カウント
			}
		}
		
		// 不良ブロックでなく && NAND Flashの全block分のBBTが書かれている ならば 終了
		if(2 <= page && bbt_page_num <= cnt){
			break;
		}
		
		// 次のblockへ
		block++;
		
		// BBTが書かれていないならば
		if(BLOCK_BBT <= block){
			// ミラーのBBTが書かれていないか？もう一回チェック
			if(td != chip->bbt_md){
				td = chip->bbt_md; // ミラーBBTデスクリプタをセット(bbt_mirror_descr です)
				block = 0;
			}
			// ミラーのBBTも書かれていないならば 終了
			else{
				printk(KERN_INFO "Nothing NAND Flash BBT\n");
				break;
			}
		}
	}
	
	// BBTが書き込まれていたら(BBTが書き込まれていないならば 何もしない)
	if(block < BLOCK_BBT){
		
		printk(KERN_INFO "Check bad block table\n");
		
		// パーティションデータをセット
		if(chip->chipsize < NAND_SIZE64){
			mtd_parts = osbourne_nand_partition_info32; /* 32Mbyte NAND Flash */
		}
		else{
			mtd_parts = osbourne_nand_partition_info64; /* 64Mbyte NAND Flash */
		}
		
		// 先頭のdummyパーティションを スキップ
		while(mtd_parts->offset != BBT_SIZE){ // ←dummyパーティションとBBTの後に mtdパーティションがあり
			mtd_parts++;                      //   最初のmtdパーティションのoffsetは BBT終端になっているので
		}
		block = BLOCK_BBT;
		
		// 各パーティションの"後天性"不良ブロック数を算出
		// ("先天性"不良ブロックは NAND Flashメーカ出荷時からのもの or 以前に復帰を試みたけど
		//  復帰出来なかったblockなので 無視)
		bad_ptt_num = 0;
		while(block < (chip->chipsize / SZ_1BLOCK)){
			next_ptt_block = block + mtd_parts->size / SZ_1BLOCK;
			cnt = 0;
			while(block < next_ptt_block){
				// BBTデータ4byteに↓16block(=(8bit*4byte)/2bit)分のデータが入っているので
				tmp = bbt_buf[block / 16]; //   "後天性"不良ブロックだけを検出したいので "先天性"不良ブロックは 00b
				tmp ^= tmp >> 1;           // ←良ブロックは 11b なので 1bit右シフトした値を XOR した値の 偶数bitが
				                           //   1bか？(つまり 01b もしくは 10bであるか？)をチェックすればいよい！
				// 全16blockが良ブロックならば 次の16blockのBBTへ
				if(!(tmp & 0x55555555)){ // ←高速にチェックしたいので32bit値で 一気に16blockをチェック！
					block += 16;
				}
				// 16block中に"後天性"不良ブロックがある場合は 1blockづつチェック
				else{	// エンディアンを変換↓
					tmp = (tmp << 24) | ((tmp & 0x0000ff00) << 8) | ((tmp & 0x00ff0000) >> 8) | (tmp >> 24);
					mask = 0x00000001LU << (2 * (block % 16));
					while(mask && block < next_ptt_block){
						if(tmp & mask){
							cnt++; // 不良ブロックをカウント
						}
						mask <<= 2;
						block++;
					}
				}
			}
			
			// "後天性"不良ブロック数が パーティション全ブロック数の 2割以上 の場合は 異常なので
			if(mtd_parts->size / SZ_1BLOCK <= cnt * 5){
				bad_ptt_num++; // 異常なパーティション数を カウント
				printk(KERN_INFO "wng : Exist Many bad blocks [%4d/%4d] : %c%s%c\n", cnt, mtd_parts->size / SZ_1BLOCK, 0x22, mtd_parts->name, 0x22);
			}
			
			// 次のパーティションへ
			block = next_ptt_block;
			mtd_parts++;
		}
		
		// "後天性"不良ブロックが多発しているパーティションがあった場合は
		// 復帰処理を行い BBTを消去して 再度 BBT作成を行う
		if(bad_ptt_num){
			
			printk(KERN_INFO "Recover bad blocks\n");
			
			for(block = BLOCK_BBT; block < chip->chipsize / SZ_1BLOCK; block++){
				// "後天性"不良ブロックならば 復帰処理を行う
				tmp = (((uint8_t *)bbt_buf)[block / 4] >> (2 * (block % 4))) & 0x03;
				if(tmp != 0 && tmp != 3){
					// 1.まず ブロック消去 を行う
					chip->erase_cmd(mtd, block*PAGE_1BLOCK);
					cnt = chip->waitfunc(mtd, chip);
					if(cnt & NAND_STATUS_FAIL){
						printk(KERN_INFO "Can't erase [%d]\n", block);
					}
					else{
						cnt = 0; // 成功ならば cnt を 0 へ
					}
					
					// 2.ブロック内の全pageへ ランダムデータを書き込み
					if(!cnt){
						mask = block; // 乱数の初期値をセット
						for(page = 0; page < PAGE_1BLOCK; page++){
							// ランダムデータを chip->buffers.databuf へセット
							tmp = (uint32_t)chip->buffers.databuf;
							for(cnt = 0; cnt < SZ_1PAGE / sizeof(uint32_t); cnt++){
								switch(cnt % 3){ // ←適当に考えた乱数発生計算です！
								case 0:
									mask ^= 0xAAAAAAAA; // 奇数bitを反転
									break;
								case 1:
									mask ^= 0x55555555; // 偶数bitを反転
									break;
								default:
									mask ^= 0xFFFFFFFF; // 全bitを反転
									mask += 0x01010101; // 全byteをインクリメント
									break;
								}
								*((uint32_t *)tmp) = mask;
								tmp += sizeof(uint32_t);
							}
							
							// 1page書き込み
							chip->oob_poi = chip->buffers.oobwbuf; // ←忘れないように！
							chip->cmdfunc(mtd, NAND_CMD_SEQIN, 0, block*PAGE_1BLOCK + page);
							chip->ecc.write_page(mtd, chip, chip->buffers.databuf);
							chip->cmdfunc(mtd, NAND_CMD_PAGEPROG, -1, -1);
							cnt = chip->waitfunc(mtd, chip);
							if(cnt & NAND_STATUS_FAIL){
								printk(KERN_INFO "Can't write [%d,%d]\n", block, page);
								break; // 失敗ならば 抜ける
							}
						}
						if(PAGE_1BLOCK <= page){
							cnt = 0; // 全page成功ならば cnt を 0 へ
						}
					}
					
					// 3.ブロック内の全pageを読み出して ECCエラー発生＆訂正不可 にならないかチェック
					if(!cnt){
						for(page = 0; page < PAGE_1BLOCK; page++){
							// 1page読み込み
							tmp = mtd->ecc_stats.failed;
							chip->oob_poi = chip->buffers.oobrbuf; // ←忘れないように！
							chip->cmdfunc(mtd, NAND_CMD_READ0, 0, block*PAGE_1BLOCK + page);
							chip->ecc.read_page(mtd, chip, chip->buffers.databuf);
							if(mtd->ecc_stats.failed != tmp){
								printk(KERN_INFO "Can't read [%d,%d]\n", block, page);
								cnt = 1; // ECCエラー発生＆訂正不可 ならば cnt を 1 へ
								break; // 抜ける
							}
						}
					}
					
					// 4.ブロック消去 しておく
					if(!cnt){
						chip->erase_cmd(mtd, block*PAGE_1BLOCK);
						cnt = chip->waitfunc(mtd, chip);
						if(cnt & NAND_STATUS_FAIL){
							printk(KERN_INFO "Can't erase [%d] 2\n", block);
						}
						else{
							cnt = 0; // 成功ならば cnt を 0 へ
						}
					}
					
					// ブロック消去 or 書き込み or 読み込み が失敗したら 0,1pageへ不良ブロックMarkを書き込み
					if(cnt){
						oob_data = chip->buffers.oobwbuf;
						memset(oob_data, 0xff, 16); oob_data[bd->offs] = 0x00;
						for(page = 0; page < 2; page++){
							chip->cmdfunc(mtd, NAND_CMD_SEQIN, mtd->writesize, block*PAGE_1BLOCK + page);
							chip->write_buf(mtd, oob_data, 16); // ココ↑に 512(= mtd->writesize)をセットで OOB書き込み
							chip->cmdfunc(mtd, NAND_CMD_PAGEPROG, -1, -1);
							chip->waitfunc(mtd, chip);
						}
					}
				}
			}
			
			// BBTを消去して MTDにBBTの再構築を行ってもらう
			for(block = 0; block < BLOCK_BBT; block++){
				chip->erase_cmd(mtd, block*PAGE_1BLOCK);
				chip->waitfunc(mtd, chip);
			}
		}
	}
	
	// NAND Flash を Inactivate
	chip->select_chip(mtd, -1);
	
	// BBT読み込みバッファを解放
	if(bbt_buf){
		kfree((void *)bbt_buf);
	}
}

MODULE_LICENSE("GPL");

