/*
 * YAFFS: Yet Another Flash File System. A NAND-flash specific file system.
 *
 * Copyright (C) 2002-2007 Aleph One Ltd.
 *   for Toby Churchill Ltd and Brightstar Engineering
 *
 * Created by Charles Manning <charles@aleph1.co.uk>
 * Nick Bane modifications flagged NCB
 * Endian handling patches by James Ng
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * 2009.07.22 H.HASEGAWA	mkyaffsimage を Linux-2.6.18 に built-in されている yaffs を
 *                     		サポートするように mkyaffsimage2 へ改造
 * 2009.07.23 H.HASEGAWA	NAND Flashの各ページ終端のOOB領域の7byte目のbit7,6へ, ブロックの
 *                      	良or不良を示すbit値 と pageStatusをを示すbit値が, Linux-2.6.18以降のyaffsから
 *                      	書き込まれるようになったが, このbitらは, 同じくOOB領域に書き込まれている
 *                      	tagsデータの objectId値 を上書きしており, objectId値が損傷を受ける yaffsの
 *                      	不具合修正に伴った修正を行う
 */

/*
 * makeyaffsimage.c 
 *
 * Makes a YAFFS file system image that can be used to load up a file system.
 */
 
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <string.h>
#include <unistd.h>
#include "yaffs_ecc.h"
#if 1 // warning対策
// Linux PC上の /usr/include/dirent.h 内にも DT_UNKNOWN らの定義があるため
// warningが発生してしまうが DT_UNKNOWN らの定義値は, dirent.h と
// ../yaffs_guts.h -> ../devextras.h で同値なので, ココで undef しておく
	#undef DT_UNKNOWN	// 0
	#undef DT_FIFO		// 1
	#undef DT_CHR		// 2
	#undef DT_DIR		// 4
	#undef DT_BLK		// 6
	#undef DT_REG		// 8
	#undef DT_LNK		// 10
	#undef DT_SOCK		// 12
	#undef DT_WHT		// 14
#endif
#include "yaffs_guts.h"


#define MAX_OBJECTS 10000

// Linux-2.6.18 へ buit-in されている yaffs対応変更 を 有効へ
#define SUPPORT_YAFFS_ON_LINUX_2_6_18

const char * mkyaffsimage_c_version = "$Id: mkyaffsimage.c,v 1.7 2003/07/16 03:00:48 charles Exp $";


typedef struct
{
	dev_t dev;
	ino_t ino;
	int   obj;
} objItem;


static objItem obj_list[MAX_OBJECTS];
static int n_obj = 0;
#ifdef SUPPORT_YAFFS_ON_LINUX_2_6_18 // 4-a
static int obj_id = 0x105; // objectId値 は 0x105 から開始すること
#else
static int obj_id = YAFFS_NOBJECT_BUCKETS + 1;
#endif

static int nObjects = 0, nDirectories = 0, nPages = 0;

static int outFile = 0;

static int error = 0;

static int convert_endian = 0;

static int obj_compare(const void *a, const void * b)
{
  objItem *oa, *ob;
  
  oa = (objItem *)a;
  ob = (objItem *)b;
  
  if(oa->dev < ob->dev) return -1;
  if(oa->dev > ob->dev) return 1;
  if(oa->ino < ob->ino) return -1;
  if(oa->ino > ob->ino) return 1;
  
  return 0;
}


static void add_obj_to_list(dev_t dev, ino_t ino, int obj)
{
	if(n_obj < MAX_OBJECTS)
	{
		obj_list[n_obj].dev = dev;
		obj_list[n_obj].ino = ino;
		obj_list[n_obj].obj = obj;
		n_obj++;
		qsort(obj_list,n_obj,sizeof(objItem),obj_compare);
		
	}
	else
	{
		// oops! not enough space in the object array
		fprintf(stderr,"Not enough space in object array\n");
		exit(2);
	}
}


static int find_obj_in_list(dev_t dev, ino_t ino)
{
	objItem *i = NULL;
	objItem test;

	test.dev = dev;
	test.ino = ino;
	
	if(n_obj > 0)
	{
		i = bsearch(&test,obj_list,n_obj,sizeof(objItem),obj_compare);
	}

	if(i)
	{
		return i->obj;
	}
	return -1;
}

#if 0 // warning対策
// NCB added 10/9/2002
static __u16 yaffs_CalcNameSum(const char *name)
{
	__u16 sum = 0;
	__u16 i = 1;
	
	__u8 *bname = (__u8 *)name;
	
	while (*bname)
	{
		sum += (*bname) * i;
		i++;
		bname++;
	}
	return sum;
}
#endif


static void yaffs_CalcECC(const __u8 *data, yaffs_Spare *spare)
{
#ifdef SUPPORT_YAFFS_ON_LINUX_2_6_18 // 4-b
	__u8 tmp;
#endif
	
	yaffs_ECCCalculate(data , spare->ecc1);
	yaffs_ECCCalculate(&data[256] , spare->ecc2);
	
#ifdef SUPPORT_YAFFS_ON_LINUX_2_6_18 // 4-b
	// ECCの0,1byteが逆なので 入れ替え
	tmp = spare->ecc1[0]; spare->ecc1[0] = spare->ecc1[1];
	spare->ecc1[1] = tmp;
	tmp = spare->ecc2[0]; spare->ecc2[0] = spare->ecc2[1];
	spare->ecc2[1] = tmp;
#endif
}

static void yaffs_CalcTagsECC(yaffs_Tags *tags)
{
	// Todo don't do anything yet. Need to calculate ecc
	unsigned char *b = ((yaffs_TagsUnion *)tags)->asBytes;
	unsigned  i,j;
	unsigned  ecc = 0;
	unsigned bit = 0;

#ifdef SUPPORT_YAFFS_ON_LINUX_2_6_18
#if 1 /* Add for bugfix damaged objectId by hiroshi.hasegawa@pixela.co.jp on 2009/07/23 */
	// tags->unusedStuff の 2bit分の領域へ, translate_spare2oob() 内にて blockStatus と pageStatus 値を示す
	// bit値がセットされるため, NAND Flash から読み出した(そのままの) tags の unusedStuff は不定値となる.
	// したがって, yaffs処理内では unusedStuff は 常に 11b(=0x3)の固定値とし 無意味な値としておく.
    if (!convert_endian)
    {
		tags->unusedStuff = 0x3;
	}
	else
    {
        // Because we're in "munged tag" mode, we have to set it manually
        b[7] &= 0xfc;
        b[7] |= 0x03;
    }
#endif
#endif

    // Clear ECC fields
    if (!convert_endian)
    {
	    tags->ecc = 0;
    }
    else
    {
        // Because we're in "munged tag" mode, we have to clear it manually
        b[6] &= 0xC0;
        b[7] &= 0x03;
    }
	
	for(i = 0; i < 8; i++)
	{
// NCB modified 20-9-02		for(j = 1; j &0x7f; j<<=1)
		for(j = 1; j &0xff; j<<=1)
		{
			bit++;
			if(b[i] & j)
			{
				ecc ^= bit;
			}
		}
	}
	
	// Write out ECC
    if (!convert_endian)
    {
        tags->ecc = ecc;
    }
    else
    {
	    // We have to munge the ECC again.
	    b[6] |= ((ecc >> 6) & 0x3F);
        b[7] |= ((ecc & 0x3F) << 2);
    }
}
static void yaffs_LoadTagsIntoSpare(yaffs_Spare *sparePtr, yaffs_Tags *tagsPtr)
{
	yaffs_TagsUnion *tu = (yaffs_TagsUnion *)tagsPtr;
	
	//yaffs_CalcTagsECC(tagsPtr);
	
	sparePtr->tagByte0 = tu->asBytes[0];
	sparePtr->tagByte1 = tu->asBytes[1];
	sparePtr->tagByte2 = tu->asBytes[2];
	sparePtr->tagByte3 = tu->asBytes[3];
	sparePtr->tagByte4 = tu->asBytes[4];
	sparePtr->tagByte5 = tu->asBytes[5];
	sparePtr->tagByte6 = tu->asBytes[6];
	sparePtr->tagByte7 = tu->asBytes[7];
}

/* This little function converts a little endian tag to a big endian tag.
 * NOTE: The tag is not usable after this other than calculating the CRC
 * with.
 */
static void little_to_big_endian(yaffs_Tags *tagsPtr)
{
    yaffs_TagsUnion * tags = (yaffs_TagsUnion* )tagsPtr; // Work in bytes.
    yaffs_TagsUnion   temp;

    memset(&temp, 0, sizeof(temp));
    // Ick, I hate magic numbers.
    temp.asBytes[0] = ((tags->asBytes[2] & 0x0F) << 4) | ((tags->asBytes[1] & 0xF0) >> 4);
    temp.asBytes[1] = ((tags->asBytes[1] & 0x0F) << 4) | ((tags->asBytes[0] & 0xF0) >> 4);
    temp.asBytes[2] = ((tags->asBytes[0] & 0x0F) << 4) | ((tags->asBytes[2] & 0x30) >> 2) | ((tags->asBytes[3] & 0xC0) >> 6);
    temp.asBytes[3] = ((tags->asBytes[3] & 0x3F) << 2) | ((tags->asBytes[2] & 0xC0) >> 6);
    temp.asBytes[4] = ((tags->asBytes[6] & 0x03) << 6) | ((tags->asBytes[5] & 0xFC) >> 2);
    temp.asBytes[5] = ((tags->asBytes[5] & 0x03) << 6) | ((tags->asBytes[4] & 0xFC) >> 2);
    temp.asBytes[6] = ((tags->asBytes[4] & 0x03) << 6) | (tags->asBytes[7] & 0x3F);
    temp.asBytes[7] = (tags->asBytes[6] & 0xFC) | ((tags->asBytes[7] & 0xC0) >> 6);

    // Now copy it back.
    tags->asBytes[0] = temp.asBytes[0];
    tags->asBytes[1] = temp.asBytes[1];
    tags->asBytes[2] = temp.asBytes[2];
    tags->asBytes[3] = temp.asBytes[3];
    tags->asBytes[4] = temp.asBytes[4];
    tags->asBytes[5] = temp.asBytes[5];
    tags->asBytes[6] = temp.asBytes[6];
    tags->asBytes[7] = temp.asBytes[7];
}

#ifdef SUPPORT_YAFFS_ON_LINUX_2_6_18 // 4-c, 4-d
static int write_chunk(__u8 *data, __u32 objId, __u32 chunkId, __u32 nBytes, unsigned serialNumber, __u8 pageStatus)
#else
static int write_chunk(__u8 *data, __u32 objId, __u32 chunkId, __u32 nBytes)
#endif
{
	yaffs_Tags t;
	yaffs_Spare s;

	error = write(outFile,data,512);
	if(error < 0) return error;

	memset(&t,0xff,sizeof (yaffs_Tags));
	memset(&s,0xff,sizeof (yaffs_Spare));
	
	t.chunkId = chunkId;
#ifdef SUPPORT_YAFFS_ON_LINUX_2_6_18 // 4-c
	t.serialNumber = serialNumber;
#else
	t.serialNumber = 0;
#endif
	t.byteCount = nBytes;
	t.objectId = objId;

    if (convert_endian)
    {
        little_to_big_endian(&t);
    }
	
	yaffs_CalcTagsECC(&t);
	yaffs_LoadTagsIntoSpare(&s,&t);
	yaffs_CalcECC(data,&s);
	
#ifdef SUPPORT_YAFFS_ON_LINUX_2_6_18 // 1, 4-d
#if 1 /* Mod for bugfix damaged objectId by hiroshi.hasegawa@pixela.co.jp on 2009/07/23 */
	// oob[5] は objectId のbit領域であり, oob[5] の bit7,5へ値がセットされると objectId値 が破損するので,
	// yaffsが未使用の unusedStuff のbit領域, oob[7] の bit1,0へ値をセット＆ゲットするように変更
	
	// tagByte7 の 1bit へ, 良ブロックフラグ値として 1 をセット
	s.tagByte7 &= 0xfc;
	s.tagByte7 |= 0x02; /* blockStatus == 'Y' ? 0: 0x02; */
	
	// tagByte7 の 0bit へ, pageStatus値 0 or 1 をセット
	s.tagByte7 |= pageStatus == 0 ? 0: 0x01;
#else
	/* ココで tagByte5 の 7,6bit を書き換えてしまうと, objectId値が
	   壊れてしまう訳だが, ↑の yaffs_CalcTagsECC() にて ECC値が
	   算出＆セットされていて, このECC値から, 壊された値は復活出来る
	   ようになっており, うまく動作しています
	   (というか, バグでない？ めちゃ, 悩まされました...) */
	
	// tagByte5 の 7bit へ, 良ブロックフラグ値として 1 をセット
	s.tagByte5 &= 0x3f;
	s.tagByte5 |= 0x80; /* blockStatus == 'Y' ? 0: 0x80; */
	
	// tagByte5 の 6bit へ, pageStatus値 0 or 1 をセット
	s.tagByte5 |= pageStatus == 0 ? 0: 0x40;
#endif
#endif
	
	nPages++;
	
	return write(outFile,&s,sizeof(yaffs_Spare));
	
}

#define SWAP32(x)   ((((x) & 0x000000FF) << 24) | \
                     (((x) & 0x0000FF00) << 8 ) | \
                     (((x) & 0x00FF0000) >> 8 ) | \
                     (((x) & 0xFF000000) >> 24))

#define SWAP16(x)   ((((x) & 0x00FF) << 8) | \
                     (((x) & 0xFF00) >> 8))
        
// This one is easier, since the types are more standard. No funky shifts here.
static void object_header_little_to_big_endian(yaffs_ObjectHeader* oh)
{
    oh->type = SWAP32(oh->type); // GCC makes enums 32 bits.
    oh->parentObjectId = SWAP32(oh->parentObjectId); // int
    oh->sum__NoLongerUsed = SWAP16(oh->sum__NoLongerUsed); // __u16 - Not used, but done for completeness.
    // name = skip. Char array. Not swapped.
    oh->yst_mode = SWAP32(oh->yst_mode);
#ifdef CONFIG_YAFFS_WINCE // WinCE doesn't implement this, but we need to just in case. 
    // In fact, WinCE would be *THE* place where this would be an issue!
    oh->notForWinCE[0] = SWAP32(oh->notForWinCE[0]);
    oh->notForWinCE[1] = SWAP32(oh->notForWinCE[1]);
    oh->notForWinCE[2] = SWAP32(oh->notForWinCE[2]);
    oh->notForWinCE[3] = SWAP32(oh->notForWinCE[3]);
    oh->notForWinCE[4] = SWAP32(oh->notForWinCE[4]);
#else
    // Regular POSIX.
    oh->yst_uid = SWAP32(oh->yst_uid);
    oh->yst_gid = SWAP32(oh->yst_gid);
    oh->yst_atime = SWAP32(oh->yst_atime);
    oh->yst_mtime = SWAP32(oh->yst_mtime);
    oh->yst_ctime = SWAP32(oh->yst_ctime);
#endif

    oh->fileSize = SWAP32(oh->fileSize); // Aiee. An int... signed, at that!
    oh->equivalentObjectId = SWAP32(oh->equivalentObjectId);
    // alias  - char array.
    oh->yst_rdev = SWAP32(oh->yst_rdev);

#ifdef CONFIG_YAFFS_WINCE
    oh->win_ctime[0] = SWAP32(oh->win_ctime[0]);
    oh->win_ctime[1] = SWAP32(oh->win_ctime[1]);
    oh->win_atime[0] = SWAP32(oh->win_atime[0]);
    oh->win_atime[1] = SWAP32(oh->win_atime[1]);
    oh->win_mtime[0] = SWAP32(oh->win_mtime[0]);
    oh->win_mtime[1] = SWAP32(oh->win_mtime[1]);
    oh->roomToGrow[0] = SWAP32(oh->roomToGrow[0]);
    oh->roomToGrow[1] = SWAP32(oh->roomToGrow[1]);
    oh->roomToGrow[2] = SWAP32(oh->roomToGrow[2]);
    oh->roomToGrow[3] = SWAP32(oh->roomToGrow[3]);
    oh->roomToGrow[4] = SWAP32(oh->roomToGrow[4]);
    oh->roomToGrow[5] = SWAP32(oh->roomToGrow[5]);
#else
    oh->roomToGrow[0] = SWAP32(oh->roomToGrow[0]);
    oh->roomToGrow[1] = SWAP32(oh->roomToGrow[1]);
    oh->roomToGrow[2] = SWAP32(oh->roomToGrow[2]);
    oh->roomToGrow[3] = SWAP32(oh->roomToGrow[3]);
    oh->roomToGrow[4] = SWAP32(oh->roomToGrow[4]);
    oh->roomToGrow[5] = SWAP32(oh->roomToGrow[5]);
    oh->roomToGrow[6] = SWAP32(oh->roomToGrow[6]);
    oh->roomToGrow[7] = SWAP32(oh->roomToGrow[7]);
    oh->roomToGrow[8] = SWAP32(oh->roomToGrow[8]);
    oh->roomToGrow[9] = SWAP32(oh->roomToGrow[9]);
    oh->roomToGrow[10] = SWAP32(oh->roomToGrow[10]);
    oh->roomToGrow[11] = SWAP32(oh->roomToGrow[11]);
#endif
}

#ifdef SUPPORT_YAFFS_ON_LINUX_2_6_18 // 4-c, 4-d
static int write_object_header(int objId, yaffs_ObjectType t, struct stat *s, int parent, const char *name, int equivalentObj, const char * alias,
                               unsigned serialNumber, __u8 pageStatus)
#else
static int write_object_header(int objId, yaffs_ObjectType t, struct stat *s, int parent, const char *name, int equivalentObj, const char * alias)
#endif
{
	__u8 bytes[512];
	
	
	yaffs_ObjectHeader *oh = (yaffs_ObjectHeader *)bytes;
	
	memset(bytes,0xff,512);
	
	oh->type = t;

	oh->parentObjectId = parent;
	
#ifdef SUPPORT_YAFFS_ON_LINUX_2_6_18
	// ファイル(ディレクトリ)名をコピー前に 0クリア しておく
	memset(oh->name, 0, sizeof(oh->name));
#endif
	strncpy(oh->name,name,YAFFS_MAX_NAME_LENGTH);
	
	
	if(t != YAFFS_OBJECT_TYPE_HARDLINK)
	{
		oh->yst_mode = s->st_mode;
#ifdef SUPPORT_YAFFS_ON_LINUX_2_6_18 // 5
		// セットのLinux上では ファイルへアクセスするのは root のみなので
		// ユーザ と グループ は 0 にしておく
		oh->yst_uid = 0;
		oh->yst_gid = 0;
#else
		oh->yst_uid = s->st_uid;
// NCB 12/9/02		oh->yst_gid = s->yst_uid;
		oh->yst_gid = s->st_gid;
#endif
		oh->yst_atime = s->st_atime;
		oh->yst_mtime = s->st_mtime;
		oh->yst_ctime = s->st_ctime;
		oh->yst_rdev  = s->st_rdev;
#ifdef SUPPORT_YAFFS_ON_LINUX_2_6_18 // 4-e
		// shadowsObject および isShrink は 0 へ
		oh->shadowsObject = 0;
		oh->isShrink = 0;
#endif
	}
	
	if(t == YAFFS_OBJECT_TYPE_FILE)
	{
#ifdef SUPPORT_YAFFS_ON_LINUX_2_6_18 // 6
		// 先頭オブジェクト・ヘッダ の場合は fileSize を 0 へ
		if(serialNumber == 1){
			oh->fileSize = 0;
		}
		else{
			oh->fileSize = s->st_size;
		}
#else
		oh->fileSize = s->st_size;
#endif
	}
	
	if(t == YAFFS_OBJECT_TYPE_HARDLINK)
	{
		oh->equivalentObjectId = equivalentObj;
	}
	
	if(t == YAFFS_OBJECT_TYPE_SYMLINK)
	{
		strncpy(oh->alias,alias,YAFFS_MAX_ALIAS_LENGTH);
	}

    if (convert_endian)
    {
        object_header_little_to_big_endian(oh);
    }
	
#ifdef SUPPORT_YAFFS_ON_LINUX_2_6_18 // 4-f
	// オブジェクト・ヘッダ の場合は nBytes(chunk内の有効データサイズ) を 0 へ
	return write_chunk(bytes, objId, 0, 0, serialNumber, pageStatus);
#else
	return write_chunk(bytes,objId,0,0xffff);
#endif
}


static int process_directory(int parent, const char *path)
{

	DIR *dir;
	struct dirent *entry;

	nDirectories++;
	
	dir = opendir(path);
	
	if(dir)
	{
		while((entry = readdir(dir)) != NULL)
		{
		
			/* Ignore . and .. */
			if(strcmp(entry->d_name,".") &&
			   strcmp(entry->d_name,".."))
 			{
 				char full_name[500];
				struct stat stats;
				int equivalentObj;
				int newObj;
				
				sprintf(full_name,"%s/%s",path,entry->d_name);
				
				lstat(full_name,&stats);
				
				if(S_ISLNK(stats.st_mode) ||
				    S_ISREG(stats.st_mode) ||
				    S_ISDIR(stats.st_mode) ||
				    S_ISFIFO(stats.st_mode) ||
				    S_ISBLK(stats.st_mode) ||
				    S_ISCHR(stats.st_mode) ||
				    S_ISSOCK(stats.st_mode))
				{
				
					newObj = obj_id++;
					nObjects++;
					
					printf("Object %d, %s is a ",newObj,full_name);
					
					/* We're going to create an object for it */
					if((equivalentObj = find_obj_in_list(stats.st_dev, stats.st_ino)) > 0)
					{
					 	/* we need to make a hard link */
					 	printf("hard link to object %d\n",equivalentObj);
					#ifdef SUPPORT_YAFFS_ON_LINUX_2_6_18 // 4-c, 4-d
						error =  write_object_header(newObj, YAFFS_OBJECT_TYPE_HARDLINK, &stats, parent, entry->d_name, equivalentObj, NULL,
						                             1, 0xff);
					#else
						error =  write_object_header(newObj, YAFFS_OBJECT_TYPE_HARDLINK, &stats, parent, entry->d_name, equivalentObj, NULL);
					#endif
					}
					else 
					{
						
						add_obj_to_list(stats.st_dev,stats.st_ino,newObj);
						
						if(S_ISLNK(stats.st_mode))
						{
					
							char symname[500];
						
							memset(symname,0, sizeof(symname));
					
							readlink(full_name,symname,sizeof(symname) -1);
						
							printf("symlink to \"%s\"\n",symname);
						#ifdef SUPPORT_YAFFS_ON_LINUX_2_6_18 // 4-c, 4-d
							error =  write_object_header(newObj, YAFFS_OBJECT_TYPE_SYMLINK, &stats, parent, entry->d_name, -1, symname,
							                             1, 0xff);
						#else
							error =  write_object_header(newObj, YAFFS_OBJECT_TYPE_SYMLINK, &stats, parent, entry->d_name, -1, symname);
						#endif

						}
						else if(S_ISREG(stats.st_mode))
						{
							printf("file, ");
						#ifdef SUPPORT_YAFFS_ON_LINUX_2_6_18 // 4-c, 4-d
							// 先頭オブジェクト・ヘッダ は serialNumber を 1, pageStatus を 0x00 へ
							error =  write_object_header(newObj, YAFFS_OBJECT_TYPE_FILE, &stats, parent, entry->d_name, -1, NULL,
							// ただし ファイルサイズ が 0byte ならば 終端オブジェクト・ヘッダ は 不要 なので pageStatus を 0xff へ
							                             1, (stats.st_size == 0) ? 0xff : 0x00);
						#else
							error =  write_object_header(newObj, YAFFS_OBJECT_TYPE_FILE, &stats, parent, entry->d_name, -1, NULL);
						#endif

							if(error >= 0)
							{
								int h;
								__u8 bytes[512];
								int nBytes;
								int chunk = 0;
								
								h = open(full_name,O_RDONLY);
								if(h >= 0)
								{
								#ifdef SUPPORT_YAFFS_ON_LINUX_2_6_18
									memset(bytes, 0, 512); // 実機のLinuxで 0クリア されているので 0クリア へ
								#else
									memset(bytes,0xff,512);
								#endif
									while((nBytes = read(h,bytes,512)) > 0)
									{
										chunk++;
									#ifdef SUPPORT_YAFFS_ON_LINUX_2_6_18 // 4-c, 4-d
										// データ・チャンク は serialNumber を 1, pageStatus を 0xff へ
										write_chunk(bytes,newObj,chunk,nBytes, 1, 0xff);
										memset(bytes, 0, 512); // 実機のLinuxで 0クリア されているので 0クリア へ
									#else
										write_chunk(bytes,newObj,chunk,nBytes);
										memset(bytes,0xff,512);
									#endif
									}
									if(nBytes < 0) 
									   error = nBytes;
									   
									printf("%d data chunks written\n",chunk);
									
								#ifdef SUPPORT_YAFFS_ON_LINUX_2_6_18 // 6
									// 終端オブジェクト・ヘッダ は serialNumber を 2, pageStatus を 0xff へ
									if(0 < stats.st_size){ // ただし ファイルサイズ が 0byte ならば 不要
										error =  write_object_header(newObj, YAFFS_OBJECT_TYPE_FILE, &stats, parent, entry->d_name, -1, NULL,
										                             2, 0xff);
									}
								#endif
								}
								else
								{
									perror("Error opening file");
								}
								close(h);
								
							}							
														
						}
						else if(S_ISSOCK(stats.st_mode))
						{
							printf("socket\n");
						#ifdef SUPPORT_YAFFS_ON_LINUX_2_6_18 // 4-c, 4-d
							error =  write_object_header(newObj, YAFFS_OBJECT_TYPE_SPECIAL, &stats, parent, entry->d_name, -1, NULL,
							                             1, 0xff);
						#else
							error =  write_object_header(newObj, YAFFS_OBJECT_TYPE_SPECIAL, &stats, parent, entry->d_name, -1, NULL);
						#endif
						}
						else if(S_ISFIFO(stats.st_mode))
						{
							printf("fifo\n");
						#ifdef SUPPORT_YAFFS_ON_LINUX_2_6_18 // 4-c, 4-d
							error =  write_object_header(newObj, YAFFS_OBJECT_TYPE_SPECIAL, &stats, parent, entry->d_name, -1, NULL,
							                             1, 0xff);
						#else
							error =  write_object_header(newObj, YAFFS_OBJECT_TYPE_SPECIAL, &stats, parent, entry->d_name, -1, NULL);
						#endif
						}
						else if(S_ISCHR(stats.st_mode))
						{
							printf("character device\n");
						#ifdef SUPPORT_YAFFS_ON_LINUX_2_6_18 // 4-c, 4-d
							error =  write_object_header(newObj, YAFFS_OBJECT_TYPE_SPECIAL, &stats, parent, entry->d_name, -1, NULL,
							                             1, 0xff);
						#else
							error =  write_object_header(newObj, YAFFS_OBJECT_TYPE_SPECIAL, &stats, parent, entry->d_name, -1, NULL);
						#endif
						}
						else if(S_ISBLK(stats.st_mode))
						{
							printf("block device\n");
						#ifdef SUPPORT_YAFFS_ON_LINUX_2_6_18 // 4-c, 4-d
							error =  write_object_header(newObj, YAFFS_OBJECT_TYPE_SPECIAL, &stats, parent, entry->d_name, -1, NULL,
							                             1, 0xff);
						#else
							error =  write_object_header(newObj, YAFFS_OBJECT_TYPE_SPECIAL, &stats, parent, entry->d_name, -1, NULL);
						#endif
						}
						else if(S_ISDIR(stats.st_mode))
						{
							printf("directory\n");
						#ifdef SUPPORT_YAFFS_ON_LINUX_2_6_18 // 4-c, 4-d
							// 先頭オブジェクト・ヘッダ は serialNumber を 1, pageStatus を 0x00 へ
							error =  write_object_header(newObj, YAFFS_OBJECT_TYPE_DIRECTORY, &stats, parent, entry->d_name, -1, NULL,
							                             1, 0x00);
						#else
							error =  write_object_header(newObj, YAFFS_OBJECT_TYPE_DIRECTORY, &stats, parent, entry->d_name, -1, NULL);
						#endif
// NCB modified 10/9/2001				process_directory(1,full_name);
							process_directory(newObj,full_name);
						#ifdef SUPPORT_YAFFS_ON_LINUX_2_6_18 // 6
							// 終端オブジェクト・ヘッダ は serialNumber を 2, pageStatus を 0xff へ
							error =  write_object_header(newObj, YAFFS_OBJECT_TYPE_DIRECTORY, &stats, parent, entry->d_name, -1, NULL,
							                             2, 0xff);
						#endif
						}
					}
				}
				else
				{
					printf(" we don't handle this type\n");
				}
			}
		}
	}
	
	return 0;

}


int main(int argc, char *argv[])
{
	struct stat stats;
#ifdef SUPPORT_YAFFS_ON_LINUX_2_6_18
	int cnt;
#endif
	
#ifdef SUPPORT_YAFFS_ON_LINUX_2_6_18
	printf("mkyaffsimage2: image building tool for YAFFS on Linux 2.6.18 built "__DATE__"\n");
#else
	printf("mkyaffsimage: image building tool for YAFFS built "__DATE__"\n");
#endif
	
	if(argc < 3)
	{
#ifdef SUPPORT_YAFFS_ON_LINUX_2_6_18
		printf("usage: mkyaffsimage2 dir image_file [convert]\n");
#else
		printf("usage: mkyaffsimage dir image_file [convert]\n");
#endif
		printf("           dir        the directory tree to be converted\n");
		printf("           image_file the output file to hold the image\n");
        printf("           'convert'  produce a big-endian image from a little-endian machine\n");
		exit(1);
	}

#ifdef SUPPORT_YAFFS_ON_LINUX_2_6_18
	// 誤った option が設定された場合には エラー終了するようにしておく
	for(cnt = 3; cnt < argc; cnt++){
		if(!strncmp(argv[cnt], "convert", strlen("convert"))){
			convert_endian = 1;
		}
		else{
        	printf("Unknown option [%s]\n", argv[cnt]);
			exit(1);
		}
	}
#else
    if ((argc == 4) && (!strncmp(argv[3], "convert", strlen("convert"))))
    {
        convert_endian = 1;
    }
#endif
    
	if(stat(argv[1],&stats) < 0)
	{
		printf("Could not stat %s\n",argv[1]);
		exit(1);
	}
	
	if(!S_ISDIR(stats.st_mode))
	{
		printf(" %s is not a directory\n",argv[1]);
		exit(1);
	}
	
	outFile = open(argv[2],O_CREAT | O_TRUNC | O_WRONLY, S_IREAD | S_IWRITE);
	
	
	if(outFile < 0)
	{
		printf("Could not open output file %s\n",argv[2]);
		exit(1);
	}
	
	printf("Processing directory %s into image file %s\n",argv[1],argv[2]);
#ifdef SUPPORT_YAFFS_ON_LINUX_2_6_18 // 7
	// 最初に書き込んでる(なぜ？) ダミーフォルダ の オブジェクト・ヘッダ は 不要
#else
	error =  write_object_header(1, YAFFS_OBJECT_TYPE_DIRECTORY, &stats, 1,"", -1, NULL);
	if(error)
#endif
	error = process_directory(YAFFS_OBJECTID_ROOT,argv[1]);
	
	close(outFile);
	
	if(error < 0)
	{
		perror("operation incomplete");
		exit(1);
	}
	else
	{
		printf("Operation complete.\n"
		       "%d objects in %d directories\n"
		       "%d NAND pages\n",nObjects, nDirectories, nPages);
	}
	
	close(outFile);
	
	exit(0);
}	

