/*
 * YAFFS: Yet another FFS. A NAND-flash specific file system. 
 * yaffs_mtdif.c  NAND mtd wrapper functions.
 *
 * Copyright (C) 2002 Aleph One Ltd.
 *   for Toby Churchill Ltd and Brightstar Engineering
 *
 * Created by Charles Manning <charles@aleph1.co.uk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */

const char *yaffs_mtdif_c_version =
    "$Id: yaffs_mtdif.c,v 1.17 2006/11/29 20:21:12 charles Exp $";

#include "yportenv.h"


#include "yaffs_mtdif.h"

#include "linux/mtd/mtd.h"
#include "linux/types.h"
#include "linux/time.h"
#include "linux/mtd/nand.h"

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,18))
static struct nand_oobinfo yaffs_oobinfo = {
	.useecc = 1,
	.eccbytes = 6,
	.eccpos = {8, 9, 10, 13, 14, 15}
};

static struct nand_oobinfo yaffs_noeccinfo = {
	.useecc = 0,
};
#endif

#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,6,17))
static inline void translate_spare2oob(const yaffs_Spare *spare, __u8 *oob)
{
	oob[0] = spare->tagByte0;
	oob[1] = spare->tagByte1;
	oob[2] = spare->tagByte2;
	oob[3] = spare->tagByte3;
	oob[4] = spare->tagByte4;
#if 1 /* Mod for bugfix damaged objectId by hiroshi.hasegawa@pixela.co.jp on 2009/07/23 */
	// oob[5] は objectId のbit領域であり, oob[5] の bit7,5へ値がセットされると objectId値 が破損するので,
	// yaffsが未使用の unusedStuff のbit領域, oob[7] の bit1,0へ値をセット＆ゲットするように変更
	oob[5] = spare->tagByte5;
	oob[6] = spare->tagByte6;
	oob[7] = spare->tagByte7 & 0xfc;
	// 不良ブロックの場合は mtd経由で mtd->block_markbad()にて 不良ブロックMark(spareエリアのspare[5]を0x00へ)を施す
	oob[7] |= spare->blockStatus == 'Y' ? 0: 0x02;
	oob[7] |= spare->pageStatus == 0 ? 0: 0x01;
#else
	oob[5] = spare->tagByte5 & 0x3f;
	oob[5] |= spare->blockStatus == 'Y' ? 0: 0x80;
	oob[5] |= spare->pageStatus == 0 ? 0: 0x40;
	oob[6] = spare->tagByte6;
	oob[7] = spare->tagByte7;
#endif
}

#if 1 /* Mod for bugfix damaged objectId by hiroshi.hasegawa@pixela.co.jp on 2009/07/23 */
static inline void translate_oob2spare(yaffs_Spare *spare, __u8 *oob, int block_isbad)
#else
static inline void translate_oob2spare(yaffs_Spare *spare, __u8 *oob)
#endif
{
	struct yaffs_NANDSpare *nspare = (struct yaffs_NANDSpare *)spare;
	spare->tagByte0 = oob[0];
	spare->tagByte1 = oob[1];
	spare->tagByte2 = oob[2];
	spare->tagByte3 = oob[3];
	spare->tagByte4 = oob[4];
#if 1 /* Mod for bugfix damaged objectId by hiroshi.hasegawa@pixela.co.jp on 2009/07/23 */
	// oob[5] は objectId のbit領域であり, oob[5] の bit7,5へ値がセットされると objectId値 が破損するので,
	// yaffsが未使用の unusedStuff のbit領域, oob[7] の bit1,0へ値をセット＆ゲットするように変更
	spare->tagByte5 = oob[5];
	spare->ecc1[0] = spare->ecc1[1] = spare->ecc1[2] = 0xff;
	spare->tagByte6 = oob[6];
	spare->tagByte7 = oob[7] | 0x03; // yaffs処理内では unusedStuff は 常に 11b(=0x3)の固定値とし 無意味な値としておく.
	// 不良ブロックか否か？はmtd経由で mtd->block_isbad() にて取得
	spare->blockStatus = (!block_isbad) ? 0xff : 'Y';
	// pageStatus は oob[7]の0bit で判断するが NAND Flashは 1bitのデータ化けが発生するのが当たり前であり
	// oob[7]の0bit の 1bitだけで判断してはヤバイ...が 実は oob[7]の0bit は NAND Flashのspareエリアのspare[4]を
	// 利用して oob[7]の0bit は正しい値を保証するように mtdドライバにて処理を行っているので大丈夫です
	// (詳細は linux-2.6.18_pro500\drivers/mtd/nand/nand_base.c 内の nand_transfer_oob(),nand_fill_oob() を参照)
	spare->pageStatus = (oob[7] & 0x01) ? 0xff : 0;
#else
	spare->tagByte5 = oob[5] == 0xff ? 0xff : oob[5] & 0x3f;
	spare->blockStatus = oob[5] & 0x80 ? 0xff : 'Y';
	spare->pageStatus = oob[5] & 0x40 ? 0xff : 0;
	spare->ecc1[0] = spare->ecc1[1] = spare->ecc1[2] = 0xff;
	spare->tagByte6 = oob[6];
	spare->tagByte7 = oob[7];
#endif
	spare->ecc2[0] = spare->ecc2[1] = spare->ecc2[2] = 0xff;

	nspare->eccres1 = nspare->eccres2 = 0; /* FIXME */
}
#endif

int nandmtd_WriteChunkToNAND(yaffs_Device * dev, int chunkInNAND,
			     const __u8 * data, const yaffs_Spare * spare)
{
	struct mtd_info *mtd = (struct mtd_info *)(dev->genericDevice);
#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,6,17))
	struct mtd_oob_ops ops;
#endif
	size_t dummy;
	int retval = 0;

	loff_t addr = ((loff_t) chunkInNAND) * dev->nDataBytesPerChunk;
#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,6,17))
	__u8 spareAsBytes[8]; /* OOB */

	if (data && !spare)
		retval = mtd->write(mtd, addr, dev->nDataBytesPerChunk,
				&dummy, data);
	else if (spare) {
		if (dev->useNANDECC) {
			translate_spare2oob(spare, spareAsBytes);
			ops.mode = MTD_OOB_AUTO;
			ops.ooblen = 8; /* temp hack */
		} else {
			ops.mode = MTD_OOB_RAW;
			ops.ooblen = YAFFS_BYTES_PER_SPARE;
		}
		ops.len = data ? dev->nDataBytesPerChunk : ops.ooblen;
		ops.datbuf = (u8 *)data;
		ops.ooboffs = 0;
		ops.oobbuf = spareAsBytes;
		retval = mtd->write_oob(mtd, addr, &ops);
	#if 1 /* Mod for bugfix damaged objectId by hiroshi.hasegawa@pixela.co.jp on 2009/07/23 */
		if (dev->useNANDECC && spare->blockStatus == 'Y' && (!mtd->block_isbad(mtd, addr)))
			mtd->block_markbad(mtd, addr); /* mtd経由で不良ブロックマーキングを行っておくこと */
	#endif
	}
#else
	__u8 *spareAsBytes = (__u8 *) spare;

	if (data && spare) {
		if (dev->useNANDECC)
			retval =
			    mtd->write_ecc(mtd, addr, dev->nDataBytesPerChunk,
					   &dummy, data, spareAsBytes,
					   &yaffs_oobinfo);
		else
			retval =
			    mtd->write_ecc(mtd, addr, dev->nDataBytesPerChunk,
					   &dummy, data, spareAsBytes,
					   &yaffs_noeccinfo);
	} else {
		if (data)
			retval =
			    mtd->write(mtd, addr, dev->nDataBytesPerChunk, &dummy,
				       data);
		if (spare)
			retval =
			    mtd->write_oob(mtd, addr, YAFFS_BYTES_PER_SPARE,
					   &dummy, spareAsBytes);
	}
#endif

	if (retval == 0)
		return YAFFS_OK;
	else
		return YAFFS_FAIL;
}

int nandmtd_ReadChunkFromNAND(yaffs_Device * dev, int chunkInNAND, __u8 * data,
			      yaffs_Spare * spare)
{
	struct mtd_info *mtd = (struct mtd_info *)(dev->genericDevice);
#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,6,17))
	struct mtd_oob_ops ops;
#endif
	size_t dummy;
	int retval = 0;

	loff_t addr = ((loff_t) chunkInNAND) * dev->nDataBytesPerChunk;
#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,6,17))
	__u8 spareAsBytes[8]; /* OOB */

	if (data && !spare) {
		retval = mtd->read(mtd, addr, dev->nDataBytesPerChunk,
				&dummy, data);
#if 1 /* Add for Return OK, when ECC recovered by hiroshi.hasegawa@pixela.co.jp on 2010/02/08 */
		/* ECCエラーが発生しても 訂正出来た際は問題ないにも関わらず -EUCLEAN が返ってくるので
		   その際は 正常終了として retval を 0 へ書き換える */
		if (retval == -EUCLEAN) {
			retval = 0;
		}
#endif
	} else if (spare) {
		if (dev->useNANDECC) {
			ops.mode = MTD_OOB_AUTO;
			ops.ooblen = 8; /* temp hack */
		} else {
			ops.mode = MTD_OOB_RAW;
			ops.ooblen = YAFFS_BYTES_PER_SPARE;
		}
		ops.len = data ? dev->nDataBytesPerChunk : ops.ooblen;
		ops.datbuf = data;
		ops.ooboffs = 0;
		ops.oobbuf = spareAsBytes;
		retval = mtd->read_oob(mtd, addr, &ops);
#if 1 /* Add for Return OK, when ECC recovered by hiroshi.hasegawa@pixela.co.jp on 2010/02/08 */
		if (retval == -EUCLEAN) {
			retval = 0;
		}
#endif
		if (dev->useNANDECC)
#if 1 /* Mod for bugfix damaged objectId by hiroshi.hasegawa@pixela.co.jp on 2009/07/23 */
			/* mtd経由で不良ブロックかどうか？問い合わせる↓ */
			translate_oob2spare(spare, spareAsBytes, mtd->block_isbad(mtd, addr));
#else
			translate_oob2spare(spare, spareAsBytes);
#endif
	}
#else
	__u8 *spareAsBytes = (__u8 *) spare;

	if (data && spare) {
		if (dev->useNANDECC) {	
			/* Careful, this call adds 2 ints */
			/* to the end of the spare data.  Calling function */
			/* should allocate enough memory for spare, */
			/* i.e. [YAFFS_BYTES_PER_SPARE+2*sizeof(int)]. */
			retval =
			    mtd->read_ecc(mtd, addr, dev->nDataBytesPerChunk,
					  &dummy, data, spareAsBytes,
					  &yaffs_oobinfo);
		} else {
			retval =
			    mtd->read_ecc(mtd, addr, dev->nDataBytesPerChunk,
					  &dummy, data, spareAsBytes,
					  &yaffs_noeccinfo);
		}
	} else {
		if (data)
			retval =
			    mtd->read(mtd, addr, dev->nDataBytesPerChunk, &dummy,
				      data);
		if (spare)
			retval =
			    mtd->read_oob(mtd, addr, YAFFS_BYTES_PER_SPARE,
					  &dummy, spareAsBytes);
	}
#endif

	if (retval == 0)
		return YAFFS_OK;
	else
		return YAFFS_FAIL;
}

int nandmtd_EraseBlockInNAND(yaffs_Device * dev, int blockNumber)
{
	struct mtd_info *mtd = (struct mtd_info *)(dev->genericDevice);
	__u32 addr =
	    ((loff_t) blockNumber) * dev->nDataBytesPerChunk
		* dev->nChunksPerBlock;
	struct erase_info ei;
	int retval = 0;

	ei.mtd = mtd;
	ei.addr = addr;
	ei.len = dev->nDataBytesPerChunk * dev->nChunksPerBlock;
	ei.time = 1000;
	ei.retries = 2;
	ei.callback = NULL;
	ei.priv = (u_long) dev;

	/* Todo finish off the ei if required */

	sema_init(&dev->sem, 0);

	retval = mtd->erase(mtd, &ei);

	if (retval == 0)
		return YAFFS_OK;
	else
		return YAFFS_FAIL;
}

int nandmtd_InitialiseNAND(yaffs_Device * dev)
{
	return YAFFS_OK;
}

